/******************************************************************************
 Symbols.c

		Symbol data manager for the SignWriter symbol editor
		
		Copyright  1995 Center for Sutton Movement Writing. All rights reserved.

 ******************************************************************************/
 
 
 
#include "Symbols.h"
#include "Global.h"
#include "Memory.h"
#include "SymEditConsts.h"





/**************************  Symbol data structures  ************************/




typedef struct {		// image 

	Point   framePt;
	Point	centerPt;
	short	numBytes;
	Byte	bitMap[1];
	
} ImageRec, *Image;



typedef struct {		// fill variation 

	ImageClass  imgClass;
	short	  	numRots;
	short	  	numFlops;
	short	  	numImgs;
	Image	  	img[1];
	
} FillVarRec, *FillVar;



typedef struct {		// symbol variation

	short	numFills;
	FillVar fill[1];
	
} SymVarRec, *SymVar;



typedef struct {		// symbol

	SymbolType  symType;
	short		numVars;
	SymVar  	var[1];
	
} SymRec, *Sym;



static Sym  	symTable[FGRID_SYMS] = {NULL};		// symbol table

static short	symAttributes;						// symbol attributes




/**************************  Private prototypes  ************************/


Image   NewImage (short numBytes);								// Allocation
FillVar NewFillVar (short numImgs);
SymVar  NewSymVar (short numFills);
Sym     NewSym (short numVars);


void    DisposImage (Image img);							    // Deallocation
void    DisposFillVar (FillVar fill, Boolean subRecs);
void    DisposSymVar (SymVar var, Boolean subRecs);
void    DisposSym (Sym sym, Boolean subRecs);


Image   IImage (void);											// Initialization
FillVar IFillVar (void);
SymVar  ISymVar (void);
Sym     ISym (void);


Image 	CopyImage (Image img);									// Duplication
FillVar CopyFillVar (FillVar fill);
SymVar  CopySymVar (SymVar var);
Sym  	CopySym (Sym sym);


void    ClearImage (Image *img);							    // Clearing
void    ClearFillVar (FillVar fill);
void    ClearSymVar (SymVar var);
//      no ClearSym

				 
void	GetByte (short *value);									// Packing support
void	GetBytes (void *base, short bytes);
void	GetWord (short *value);
void	PutByte (short value);
void	PutBytes (void *base, short bytes);
void	PutWord (short value);
void	PutWordAt (short value, unsigned short inx);		 		 
				 
				 
void	PackImage (Image img, unsigned short inx);				// Packing
void	PackFillVar (FillVar fillVar, unsigned short inx);
void	PackSymVar (SymVar symVar, unsigned short inx);
void	PackSym (Sym sym);


void	UnpackImage (Image *img);								// Unpacking
void	UnpackFillVar (FillVar *fillVar);
void	UnpackSymVar (SymVar *symVar);
void	UnpackSym (Sym *sym);

 
 
			

/*****************************  Allocation  *****************************/


void FailMemError (void) 

{
	if (MemError() != 0) {
	
		printf("\n *** Memory error ***");
		exit(0);
	}
}



/***
 * NewImage
 *
 ***/
 
Image NewImage (short numBytes)

{
	short   newSize = sizeof(ImageRec) + ((numBytes - 1) * sizeof(Byte));
	Image	newImg;
	
	newImg  = (Image) NewPtrClear(newSize);
	FailMemError();
	
	newImg->numBytes = numBytes;
	return newImg;
}
	  


/***
 * NewFillVar
 *
 ***/
 
FillVar NewFillVar (short numImgs)

{
	short   newSize = sizeof(FillVarRec) + ((numImgs - 1) * sizeof(Image));
	FillVar	newFill;
	
	newFill = (FillVar) NewPtrClear(newSize);
	FailMemError();
	
	newFill->numImgs = numImgs;
	return newFill;
}



/***
 * NewSymVar
 *
 ***/
 	
SymVar NewSymVar (short numFills)

{
	short   newSize = sizeof(SymVarRec) + ((numFills - 1) * sizeof(FillVar));
	SymVar	newVar;
	
	newVar  = (SymVar) NewPtrClear(newSize);
	FailMemError();
	
	newVar->numFills = numFills;
	return newVar;
}



/***
 * NewSym
 *
 ***/
 
Sym NewSym (short numVars)

{
	short newSize = sizeof(SymRec) + ((numVars - 1) * sizeof(SymVar));
	Sym	  newSym;
	
	newSym  = (Sym) NewPtrClear(newSize);
	FailMemError();
	
	newSym->numVars = numVars;
	return newSym;
}



/***
 * NewSymbol
 *
 *	Initialize an empty symbol entry
 *  for the specified symbol code.
 *
 ***/
 
void NewSymbol (short symCode)

{	
	symTable[symCode] = NULL;	
}




/*****************************  Deallocation  *****************************/



/***
 * DisposImage
 *
 ***/
 
void  DisposImage (Image img)

{
	if (img == NULL) return;
	
	DisposPtr(img);
	FailMemError();	
}



/***
 * DisposFillVar
 *
 ***/
 
 void  DisposFillVar (FillVar fill, Boolean subRecs)

{
	short x;
	
	if (subRecs)
		for (x = 0; x < fill->numImgs; x++)
			DisposImage(fill->img[x]);
	
	DisposPtr(fill);
	FailMemError();
}



/***
 * DisposSymVar
 *
 ***/
 
void  DisposSymVar (SymVar var, Boolean subRecs)

{
	short x;
	
	if (subRecs)
		for (x = 0; x < var->numFills; x++)
			DisposFillVar(var->fill[x], TRUE);
	
	DisposPtr(var);
	FailMemError();
}



/***
 * DisposSym
 *
 ***/
 
void  DisposSym (Sym sym, Boolean subRecs)

{
	short x;
	
	if (sym == NULL) return;

	if (subRecs)
		for (x = 0; x < sym->numVars; x++)
			DisposSymVar(sym->var[x], TRUE);
	
	DisposPtr(sym);
	FailMemError();
}



/***
 * DisposSymbol
 *
 ***/
 
void  DisposSymbol (short symCode)

{	
	DisposSym(symTable[symCode], TRUE);	
}




/**************************  Initialization  **************************/



/***
 * IImage
 *
 *	Return a value indicating an empty
 *  image.  The actual image record is
 *  created when an image is saved.
 *
 ***/

Image IImage (void)

{
	return NULL;
}



/***
 * IFillVar
 *
 *	Create a fill variation record and
 *  initialize it to contain a single
 *  (empty) image.
 *
 ***/

FillVar IFillVar (void)

{
	FillVar fillVar = NewFillVar(1);
	
	fillVar->imgClass = class1;
	fillVar->numRots  = 0;
	fillVar->numFlops = 0;
	fillVar->img[0]   = IImage();
	
	return fillVar;
}



/***
 * ISymVar
 *
 *	Create the specified symbol variation
 *  and initialize it to contain a single
 *  fill variation.
 *
 ***/
 
SymVar ISymVar (void)

{
	SymVar symVar = NewSymVar(1);	
	
	symVar->fill[0] = IFillVar();
	
	return symVar;	
}



/***
 *  ISym
 *
 *	Create the specified symbol record
 *  and initialize it to contain a single
 *  symbol variation.
 *
 ***/
 
Sym ISym (void)

{	
	Sym sym = NewSym(1);

	sym->symType = other;	
	sym->var[0] = ISymVar();
	
	return sym;	
}



/***
 * ISymbol
 *
 *	Create a symbol entry for the specified
 *  symcode and initialize it to contain a 
 *  single symbol variation.
 *
 ***/
 
void ISymbol (short symCode)

{	
	symTable[symCode] = ISym();	
}




/*****************************  Duplication  *****************************/


/***
 * CopyImage
 *
 ***/
 
Image CopyImage (Image img)

{
	Image  newImg;
	
	if (img == NULL) return NULL;
	
	newImg = NewImage(img->numBytes);
	
	newImg->framePt  = img->framePt;	
	newImg->centerPt = img->centerPt;	
		
	BlockMove(&img->bitMap[0], &newImg->bitMap[0], img->numBytes);
	
	return newImg;
}



/***
 * CopyFillVar
 *
 ***/
 
FillVar CopyFillVar (FillVar fill)

{
	FillVar	  newFillVar = NewFillVar(fill->numImgs);
	short 	  x;

	newFillVar->imgClass = fill->imgClass;	
	newFillVar->numFlops = fill->numFlops;	
	newFillVar->numRots  = fill->numRots;	

	for (x = 0; x < fill->numImgs; x++)
		newFillVar->img[x] = CopyImage(fill->img[x]);
	
	return newFillVar;
}



/***
 * CopySymVar
 *
 ***/
 	
SymVar CopySymVar (SymVar var)

{
	SymVar	newSymVar = NewSymVar(var->numFills);
	short 	x;

	for (x = 0; x < var->numFills; x++)
		newSymVar->fill[x] = CopyFillVar(var->fill[x]);
	
	return newSymVar;
}



/***
 * CopySym
 *
 ***/
	
Sym CopySym (Sym sym)

{
	Sym		newSym;
	short 	x;
	
	if (sym == NULL) return NULL;
	
	newSym = NewSym(sym->numVars);

	newSym->symType = sym->symType;	

	for (x = 0; x < sym->numVars; x++)
		newSym->var[x] = CopySymVar(sym->var[x]);
	
	return newSym;
}




/*****************************  Clearification  *****************************/


/***
 * ClearImage
 *
 ***/
 
void  ClearImage (Image *img)

{
	DisposImage(*img);
	*img = NULL;
}



/***
 * ClearFillVar
 *
 ***/
 
 void  ClearFillVar (FillVar fill)

{
	short x;
	
	for (x = 0; x < fill->numImgs; x++)
		ClearImage(&fill->img[x]);
}



/***
 * ClearSymVar
 *
 ***/
 
void  ClearSymVar (SymVar var)

{
	short x;
		
	for (x = 0; x < var->numFills; x++)
		ClearFillVar(var->fill[x]);
}



// no ClearSymbol




/***********************  Attribute operations  ************************/


/***
 * GetSymType
 *
 ***/
 
void GetSymType (SymbolType *symType)

{
	short	typeVal = (symAttributes >> 6) & 7;
	
	switch (typeVal) {
	
		case 0:
		case 1:
			*symType = face;
			break;
			
		case 2:
			*symType = hand;
			break;
		
		case 3:
		case 7:
			*symType = movement;
			break;
			
		case 4:
			*symType = shoulder;
			break;
			
		case 5:
			*symType = contact;
			break;
			
		case 6:
			*symType = other;
			break;
			
	}

}



/***
 * GetFillVars
 *
 ***/
 
void GetFillVars (short *numFills)

{
	*numFills = symAttributes & 7;
}



/***
 * GetClass
 *
 ***/	
 			   
void GetClass (ImageClass *imgClass)

{
	short	numImgs = (symAttributes >> 3) & 7;
	
	switch (numImgs) {
	
		case 1:
			*imgClass = class1;
			break;
			
		case 2:
			*imgClass = class2;
			break;
		
		case 3:
			*imgClass = class3A;
			break;
			
		case 4:
			*imgClass = class4;
			break;
			
		case 5:
			*imgClass = class5;
			break;
			
		otherwise:
			printf("\n *** Sym attribute error ***");
			exit(0);
	}
}			    



/***
 * GetRotations
 *
 ***/
 
void  GetRotations (short *numRots)

{
	*numRots = (symAttributes >> 9) & 7;
}	



/***
 * GetFlops
 *
 ***/
 
void  GetFlops (short *numFlops)

{
	*numFlops = (symAttributes >> 12) & 1;
}		
	
				

/***
 * GetImages
 *
 ***/		
 		
void  GetImages (ImageClass imgClass, short *numImgs)

{
	switch (imgClass) {
	
		case class1:
			*numImgs = 1;
			break;
			
		case class2:
			*numImgs = 2;
			break;	
			
		case class3A:
		case class3B:
			*numImgs = 3;
			break;	
			
		case class4:
			*numImgs = 4;
			break;
			
		case class5:
			*numImgs = 5;
			break;
		
		default:
			*numImgs = 1;
			break;
	}	
}

/**************************  Symbol editing  ****************************/
				 				  

/***
 * GetSymbolRec
 *
 ***/
 
void  GetSymbolRec (short symCode,
					void **symbolRec)

{
	*symbolRec = (void *) symTable[symCode];
}



/***
 * SetSymbolRec
 *
 ***/	
 			  
void  SetSymbolRec (short symCode,
					void *symbolRec)

{
	symTable[symCode] = (Sym) symbolRec;
}

			

/***
 * CopySymbolRec
 *
 ***/
 
void  CopySymbolRec (void *symbolRec, void **copyRec)

{
	*copyRec = (void *) CopySym((Sym) symbolRec);
}



// no ClearSymbolRec



/***
 * DisposSymbolRec
 *
 ***/
 				  
void  DisposSymbolRec (void *symbolRec)

{
	DisposSym((Sym) symbolRec, TRUE);
}




/***********************  Sym variation editing  ************************/
				 				  

/***
 * GetSymVarRec
 *
 ***/
 
void  GetSymVarRec (short symCode, short symVar, 
					void **symVarRec)

{
	*symVarRec = (void *) symTable[symCode]->var[symVar-1];
}



/***
 * SetSymVarRec
 *
 ***/	
 			  
void  SetSymVarRec (short symCode, short symVar,
					void *symVarRec)

{
	symTable[symCode]->var[symVar-1] = (SymVar) symVarRec;
}

			

/***
 * CopySymVarRec
 *
 ***/

void  CopySymVarRec (void *symVarRec, void **copyRec)

{
	*copyRec = (void *) CopySymVar((SymVar) symVarRec);
}



/***
 * ClearSymVarRec
 *
 ***/
 
void  ClearSymVarRec (void **symVarRec)

{
	ClearSymVar((SymVar) *symVarRec);
}



/***
 * DisposSymVarRec
 *
 ***/
 				  
void  DisposSymVarRec (void *symVarRec)

{
	DisposSymVar((SymVar) symVarRec, TRUE);
}


	 	
				 	
/***********************  Fill variation editing  ************************/
				 	

/***
 * GetFillVarRec
 *
 ***/
 
void  GetFillVarRec (short symCode, short symVar, short fillVar,
					void **fillVarRec)

{
	*fillVarRec = (void *) symTable[symCode]->var[symVar-1]->fill[fillVar-1];
}



/***
 * SetFillVarRec
 *
 ***/	
 			  
void  SetFillVarRec (short symCode, short symVar, short fillVar,
					void *fillVarRec)

{
	symTable[symCode]->var[symVar-1]->fill[fillVar-1] = (FillVar) fillVarRec;
}

			

/***
 * CopyFillVarRec
 *
 ***/

void  CopyFillVarRec (void *fillVarRec, void **copyRec)

{
	*copyRec = (void *) CopyFillVar((FillVar) fillVarRec);
}



/***
 * ClearFillVarRec
 *
 ***/
 				  
void  ClearFillVarRec (void **fillVarRec)

{
	ClearFillVar((FillVar) *fillVarRec);
}



/***
 * DisposFillVarRec
 *
 ***/
 				  
void  DisposFillVarRec (void *fillVarRec)

{
	DisposFillVar((FillVar) fillVarRec, TRUE);
}




/***************************  Image editing  ****************************/


/***
 * GetImageRec
 *
 ***/
 
void  GetImageRec (short symCode, short symVar, short fillVar, short image,
				   void **imageRec)
				   
{
	*imageRec = (void *) symTable[symCode]->var[symVar-1]->fill[fillVar-1]->img[image-1];
}



/***
 * SetImageRec
 *
 ***/
 				  
void  SetImageRec (short symCode, short symVar, short fillVar, short image,
				   void *imageRec)
				   
{
	symTable[symCode]->var[symVar-1]->fill[fillVar-1]->img[image-1] = (Image) imageRec;
}		

		

/***
 * CopyImageRec
 *
 ***/

void  CopyImageRec (void *imageRec, void **copyRec)

{
	*copyRec = (void *) CopyImage((Image) imageRec);
}



/***
 * ClearImageRec
 *
 ***/
 				  
void  ClearImageRec (void **imageRec)

{
	ClearImage((Image *) imageRec);
}



/***
 * DisposImageRec
 *
 ***/
 				  
void  DisposImageRec (void *imageRec)

{
	DisposImage((Image) imageRec);
}




/************************   Symbol packing & unpacking  ************************/



static	unsigned short  bufIndex;				// symbol buffer index

static	Handle			symBuffer;				// symbol buffer handle




/***
 * GetByte
 *
 ***/
 
void  GetByte (short *value)

{
	Byte  b;
	
	BlockMove(*symBuffer+bufIndex, &b, 1);
	bufIndex++;
	
	*value = b;
}



/***
 * GetBytes
 *
 ***/
 
void  GetBytes (void *base, short bytes)

{
	BlockMove(*symBuffer+bufIndex, base, bytes);
	bufIndex += bytes;
}



/***
 * GetWord
 *
 ***/
 
void  GetWord (short *value)

{
	BlockMove(*symBuffer+bufIndex, value, 2);
	bufIndex += 2;
}



/***
 * PutByte
 *
 ***/
 
void  PutByte (short value)

{
	Byte  b = value;
	
	BlockMove(&b,*symBuffer+bufIndex, 1);
	bufIndex++;
}



/***
 * PutBytes
 *
 ***/
 
void  PutBytes (void *base, short bytes)

{
	BlockMove(base, *symBuffer+bufIndex, bytes);
	bufIndex += bytes;
}



/***
 * PutWord
 *
 ***/
 
void  PutWord (short value)

{
	BlockMove(&value, *symBuffer+bufIndex, 2);
	bufIndex += 2;
}



/***
 * PutWordAt
 *
 ***/
 
void  PutWordAt (short value, unsigned short inx)

{
	BlockMove(&value, *symBuffer+inx, 2);
}




/*****************************   Symbol packing   *****************************/


/***
 * PackImage
 *
 ***/
 
void  PackImage (Image img, unsigned short inx)

{
	Point	framePt  = img->framePt;
	Point	centerPt = img->centerPt;
	short	numBytes = img->numBytes;
	short	x;
	
	
	PutWordAt(bufIndex, inx);
	
	PutByte(framePt.v);
	PutByte(framePt.h);
	PutByte(centerPt.v);
	PutByte(centerPt.h);
	
	PutWord(numBytes);

	PutBytes(&img->bitMap, numBytes);
	
}
	  


/***
 * PackFillVar
 *
 ***/
 
void  PackFillVar (FillVar fillVar, unsigned short inx)

{
	short			imgClass = fillVar->imgClass;
	short			numRots  = fillVar->numRots;
	short			numFlops = fillVar->numFlops;
	short			numImgs  = fillVar->numImgs;
	unsigned short	img[SGRID_ROWS];
	short			x;
	
	
	PutWordAt(bufIndex, inx);
	
	PutByte(imgClass);
	PutByte(numRots);
	PutByte(numFlops);
	
	PutByte(numImgs);

	for (x = 0; x < numImgs; x++) {
		img[x] = bufIndex;
		PutWord(0);
	}
	
	for (x = 0; x < numImgs; x++)
		if (fillVar->img[x] != NULL)					// skip empty images
			PackImage(fillVar->img[x], img[x]);
	
}



/***
 * PackSymVar
 *
 ***/
 	
void  PackSymVar (SymVar symVar, unsigned short inx)

{
	short			numFills = symVar->numFills;
	unsigned short	fill[SGRID_COLS];
	short			x;
	
	
	PutWordAt(bufIndex, inx);

	PutByte(numFills);
	
	PutByte(0);		// word-align filler

	for (x = 0; x < numFills; x++) {
		fill[x] = bufIndex;
		PutWord(0);
	}
	
	for (x = 0; x < numFills; x++)
		PackFillVar(symVar->fill[x], fill[x]);

}



/***
 * PackSym
 *
 ***/
 
void PackSym (Sym sym)

{
	short			symType = sym->symType;
	short			numVars = sym->numVars;
	unsigned short  var[SGRID_VARS];
	short			x;
	
		
	PutByte(symType);
	
	PutByte(numVars);
	
	for (x = 0; x < numVars; x++) {
		var[x] = bufIndex;
		PutWord(0);
	}
	
	for (x = 0; x < numVars; x++)
		PackSymVar(sym->var[x], var[x]);

}



/***
 * PackSymbol
 *
 ***/
 
void  PackSymbol (short symCode, Handle symBuf, unsigned short *symBytes)

{
	if (symTable[symCode] == NULL) {				/* empty sym */
		*symBytes = 0;
		return;
	}
	
	symBuffer = symBuf;
	bufIndex  = 0;
	
	PackSym(symTable[symCode]);
	
	*symBytes = bufIndex;
}




/***************************   Symbol unpacking   ******************************/


/***
 * UnpackImage
 *
 ***/
 
void  UnpackImage (Image *img)

{
	short	imgBytes;
	short	imgHeight;
	short	imgWidth;
	short	macWidth;
	short	byteVal;
	Point	framePt;
	Point	centerPt;
	short	inx;
	short	x, y;
	
	
	GetByte(&imgBytes);					// image length (bytes)
	
	GetByte(&imgWidth);					// image width & height (bytes)
	imgWidth = imgWidth & 31;
	imgHeight = imgBytes / imgWidth;
	
	macWidth = imgWidth;				// Mac image width (bytes)
	if ((imgWidth % 2) != 0)
		macWidth++;
	
	GetByte(&byteVal);					// skip arrowhead coords
	GetByte(&byteVal);
	
	GetByte(&centerPt.h);				// center & frame points
	
	GetByte(&byteVal);	
	centerPt.v = (byteVal & 31) * 2;						// (PC->Mac scaling)
	
	framePt.h = (imgWidth * 8) - ((byteVal >> 5) & 7);
	framePt.v = imgHeight * 2;								// (PC->Mac scaling)

	*img = NewImage(imgHeight * macWidth * 2);				// (PC->Mac scaling)
	
	(*img)->framePt  = framePt;
	(*img)->centerPt = centerPt;

	inx = 0;
	
	for (y = 0; y < imgHeight; y++) {
	
		for (x = 0; x < (imgWidth); x++) {
			
			GetByte(&byteVal);	
			(*img)->bitMap[inx] = byteVal;
			inx++;	
		}													
		
		if (macWidth > imgWidth) {							// Mac X-bytes must be even
		
			(*img)->bitMap[inx] = 0;
			inx++;	
		}				
															// (PC->Mac scaling)
															
		BlockMove(&((*img)->bitMap[inx-macWidth]), &((*img)->bitMap[inx]), macWidth);
		inx += macWidth;
	}
	
}
	  


/***
 * UnpackFillVar
 *
 ***/
 
void  UnpackFillVar (FillVar *fillVar)

{
	ImageClass		imgClass;
	short			numRots;
	short			numFlops;
	short			numImgs;
	short			x;
	
	
	GetClass(&imgClass);
	GetRotations(&numRots);
	GetFlops(&numFlops);
	
	GetImages(imgClass, &numImgs);
		
	*fillVar = NewFillVar(numImgs);	
	
	(*fillVar)->imgClass = imgClass;
	(*fillVar)->numRots  = numRots;
	(*fillVar)->numFlops = numFlops;

	for (x = 0; x < numImgs; x++)
		UnpackImage(&((*fillVar)->img[x]));
		
}



/***
 * UnpackSymVar
 *
 ***/
 	
void  UnpackSymVar (SymVar *symVar)

{
	short			numFills;
	short			x;
			
	GetFillVars(&numFills);
	
	*symVar = NewSymVar(numFills);	
	
	for (x = 0; x < numFills; x++)
		UnpackFillVar(&((*symVar)->fill[x]));
		
}



/***
 * UnpackSym
 *
 ***/
 
void UnpackSym (Sym *sym)

{
	SymbolType		symType;
	short			numVars;	
	
	GetSymType(&symType);
	numVars = 1;
		
	*sym = NewSym(numVars);	
	
	(*sym)->symType = symType;
	
	UnpackSymVar(&((*sym)->var[0]));

}



/***
 * UnpackSymbol
 *
 ***/
 
void  UnpackSymbol (short symCode, Handle symBuf, short symBytes, short symAttrs)

{
	if (symBytes == 0) {				/* empty sym */
		symTable[symCode] = NULL;
		return;
	}
	
	symBuffer = symBuf;
	bufIndex  = 0;
		
	symAttributes = symAttrs;

	UnpackSym(&symTable[symCode]);
	
}


