/******************************************************************************
 CImgWndImage.c

							The ImageTask Classes
		
	SUPERCLASS = CImgWndTask
	
	Copyright  1995 The Center for Sutton Movement Writing.  All Rights Reserved.

 ******************************************************************************/


#include "CImgWndImage.h"
#include <Global.h>
#include "Commands.h"
#include "CBureaucrat.h"
#include "CDirectorOwner.h"
#include "CImgWndPane.h"
#include "SymEditConsts.h"
#include "TBUtilities.h"





/******************************************************************************
 IImageTask

		Initialize an ImageTask object
 ******************************************************************************/

void CImageTask::IImageTask (short aNameIndex, CImgWndPane *aPane, CBitMap *aBitMap)

{	
	done = FALSE;
									
	TRY
	{	
		CImgWndTask::IImgWndTask(aNameIndex, aPane, aBitMap);
	
		srcMap = NULL;
		destMap = NULL;
		
		srcMap = (byteMapHandle) NewHandle(sizeof(byteMap));
		destMap = (byteMapHandle) NewHandleClear(sizeof(byteMap));	// init dest	
	} 
	CATCH
	{
		Dispose();
	}
	ENDTRY;
}


/******************************************************************************
 Dispose
 ******************************************************************************/

void CImageTask::Dispose (void)
{
	if (srcMap != NULL) DisposHandle(srcMap);
	if (destMap != NULL) DisposHandle(destMap);
	
	inherited::Dispose();
}


/******************************************************************************
 Round
 ******************************************************************************/

short CImageTask::Round (float val)
{
	if (val >= 0)
		return val + 0.5;
	else
		return val - 0.5;
}


/******************************************************************************
 AlertNoRoom
 
 		Display dialog alert box indicating that there's not enough 
 		room around the current image to perform the specified transform.
 ******************************************************************************/

void CImageTask::AlertNoRoom (void)
{
	short	result;
	
	PositionDialog('ALRT', ALRTnoRoom);
	result = Alert(ALRTnoRoom, NULL);		
}


/******************************************************************************
 Unpack
 
 		Unpack bitmap structure into byte map array (for image transform)
 ******************************************************************************/

void CImageTask::Unpack (BitMapPtr src, byteMap dest)
{
	short  sinx, dinx, bit, byte, bytes;

	dinx = 0;
	bytes = src->rowBytes * IMAP_HEIGHT;
	
	for (sinx = 0; sinx < bytes; sinx++) {
		byte = src->baseAddr[sinx];
		for (bit = 7; bit >=0; bit--) {
			((Byte *) dest)[dinx] = (byte >> bit) & 1;
			dinx++;
		}
	}
}


/******************************************************************************
 Pack
 
 		Pack byte map array into bitmap structure
 ******************************************************************************/

void CImageTask::Pack (byteMap src, BitMapPtr dest)
{
	short  sinx, dinx, bit, byte, bytes;
	
	sinx = 0;
	bytes = dest->rowBytes * IMAP_HEIGHT;
	
	for (dinx = 0; dinx < bytes; dinx++) {
		for (bit = 0; bit < 8; bit++) {
			byte = (byte << 1) | ((Byte *)src)[sinx];
			sinx++;
		}
		dest->baseAddr[dinx] = byte;
	}
}


/******************************************************************************
 Fill
 
 		Fill in gaps left by imperfect transform
 ******************************************************************************/

void CImageTask::Fill (byteMap map)
{
	short  i, j;
	
	for (i = 0; i < IMAP_HEIGHT; i++) {
		for (j = 0; j < (IMAP_WIDTH - 2); j++) {
			if (map[i][j] == 1) {
				if (map[i][j+1] == 0 && map[i][j+2] == 1) {
					map[i][j+1] = 1;
				}
			}
		}
	}
}


/******************************************************************************
 Fill2
 
 		Fill in gaps left by imperfect transform  (Enlarge)
 ******************************************************************************/

void CImageTask::Fill2 (byteMap map)
{
	short  i, j;
	
	for (i = 0; i < (IMAP_HEIGHT - 2); i++) {
		for (j = 0; j < IMAP_WIDTH; j++) {
			if (map[i][j] == 1) {
				if (map[i+1][j] == 0 && map[i+2][j] == 1) {
					map[i+1][j] = 1;
				}
			}
		}
	}				
}	
	
		
/******************************************************************************
 Do 
		Perform image command using specified transform
 ******************************************************************************/

void CImageTask::Do (void)

{
	LongRect	bounds;
	short  		i, j;
	byteMap		*src, *dest;
	
	if (!done) {
		
		done = TRUE;
		
		itsPane->Prepare();
		
		itsBitMap->GetBounds(&bounds);
		SaveForUndo(&bounds);
		
		HLock(srcMap);
		HLock(destMap);	
		
		src  = *srcMap;
		dest = *destMap;
		
		Unpack(itsBitMap->macBitMap, *src);
		itsPane->PaneToBitMapPt(&itsPane->itsCenter, &centerPt);

		if (Transform(*src, *dest)) {
		
			Pack(*dest, itsBitMap->macBitMap);	
			itsPane->BitMapToPanePt(&centerPt, &itsPane->itsCenter);
		}
		
		HUnlock(srcMap);
		HUnlock(destMap);	
		
		itsPane->Update(itsBitMap->macBitMap, &bounds);
	}
}
	
		
/******************************************************************************
 Transform

		Transform image
 ******************************************************************************/

Boolean CImageTask::Transform (byteMap src, byteMap dest)
{
	return FALSE;		// null transform
}

	
	
	
			/*** Classes for Individual Image Commands ***/
			
			
/******************************************************************************
 Transform {InvertVert}

		Invert image about the horizontal axis
 ******************************************************************************/

Boolean CInvertVert::Transform (byteMap src, byteMap dest)

{
	short  	i, j;
	LongPt	pt;

	for (i = 0; i < IMAP_HEIGHT; i++) {
		for (j = 0; j < IMAP_WIDTH; j++) {
			dest[i][j] = src[IMAP_HEIGHT - 1 - i][j];		
		}
	}
	
	// Update fix point
	centerPt.v = IMAP_HEIGHT - centerPt.v;
	
	return TRUE;
}

			
/******************************************************************************
 Transform {InvertHorz}

		Invert image about the vertical axis
 ******************************************************************************/

Boolean CInvertHorz::Transform (byteMap src, byteMap dest)

{
	short  	i, j;
	LongPt	pt;

	for (i = 0; i < IMAP_HEIGHT; i++) {
		for (j = 0; j < IMAP_WIDTH; j++) {
			dest[i][j] = src[i][IMAP_WIDTH - 1 - j];		
		}
	}
	
	// Update fix point
	centerPt.h = IMAP_WIDTH - centerPt.h;


	return TRUE;
}

			
/******************************************************************************
 Transform {Rotate45}

		Rotate image 45 degrees counter-clockwise about its center point
 ******************************************************************************/

Boolean CRotate45::Transform (byteMap src, byteMap dest)

{
	short  	i, j;
	short	x, y, dx, dy;
	short	xfix, yfix;

	xfix = centerPt.h;
	yfix = centerPt.v;
			
	for (i = 0; i < IMAP_HEIGHT; i++) {
		for (j = 0; j < IMAP_WIDTH; j++) {	
			if (src[i][j] != 0) {
				
				x = j - xfix;
				y = i - yfix;
				
				dx = Round( -(0.2929 * x) + (0.7071 * y) );
				dy = Round( -(0.7071 * x) - (0.2929 * y) );
				  
				if ( (i + dy) >= 0 && (i + dy) < IMAP_HEIGHT &&
					 (j + dx) >= 0 && (j + dx) < IMAP_WIDTH ) {
					 
					 dest[i + dy][j + dx] = 1;
					 
				} else {	
					AlertNoRoom();
					return FALSE;
				}
			}
		}
	}
	
	Fill(dest);
	
	return TRUE;
}

		
/******************************************************************************
 Transform {Rotate90}

		Rotate image 90 degrees counter-clockwise about its center point
 ******************************************************************************/

Boolean CRotate90::Transform (byteMap src, byteMap dest)

{
	short  	i, j;
	short	x, y, nx, ny;
	short	xfix, yfix;
	
	xfix = centerPt.h;
	yfix = centerPt.v;
	
	for (i = 0; i < IMAP_HEIGHT; i++) {
		for (j = 0; j < IMAP_WIDTH; j++) {
			if (src[i][j] == 1) {
				
				x = j - xfix;
				y = i - yfix;
				
				nx = y + xfix;
				ny = yfix - x;
		
				if ( ny >= 0 && ny < IMAP_HEIGHT &&
					 nx >= 0 && nx < IMAP_WIDTH ) {
					 
					 dest[ny][nx] = 1;
					 
				} else {	
					AlertNoRoom();
					return FALSE;
				}
			}
		}
	}
	
	Fill(dest);
	
	return TRUE;
}


		
/******************************************************************************
 Transform {Enlarge50}

		Enlarge image 50 percent about its center point
 ******************************************************************************/

Boolean CEnlarge50::Transform (byteMap src, byteMap dest)

{
	float	scale = 1.5;
	
	short  	i, j;
	short	x, y, nx, ny;
	short	xfix, yfix;
	
	xfix = centerPt.h;
	yfix = centerPt.v;
	
	for (i = 0; i < IMAP_HEIGHT; i++) {
		for (j = 0; j < IMAP_WIDTH; j++) {
			if (src[i][j] == 1) {
				
				x = j - xfix;
				y = yfix - i;
				
				nx = Round(x * scale);
				ny = Round(y * scale);
				
				nx = nx + xfix;
				ny = yfix - ny;
				
				if ( ny >= 0 && ny < IMAP_HEIGHT &&
					 nx >= 0 && nx < IMAP_WIDTH ) {
					 
					 dest[ny][nx] = 1;
					 
				} else {	
					AlertNoRoom();
					return FALSE;
				}
			}
		}
	}
	
	Fill(dest);
	Fill2(dest);
	
	return TRUE;
}
