/*****
 * CSymEditApp.c
 *
 *	Application methods for the SignWriter symbol editor.
 *
 *  Copyright  1995 The Center for Sutton Movement Writing.  All rights reserved.
 *
 *****/


#include "Commands.h"
#include "CBartender.h"
#include "CBitMap.h"
#include "CSymEditApp.h"
#include "CFileWindow.h"
#include "CFileWndGrid.h"
#include "SymEditConsts.h"
#include "Symbols.h"
#include "TBUtilities.h"




/*** Global Variables ***/


extern	OSType		 gSignature;
extern	CBureaucrat *gGopher;
extern  CBartender  *gBartender;
extern long			 gSleepTime;		/* Max time between events			*/


CFileWndGrid	*gFileWind;				/* Window updating   				*/


void			*gClipSym;				/* File window editing				*/
void			*gScratchSym;


void			*gClipRec;				/* Sym window editing				*/
void			*gScratchRec;

short			 gClipItem;
short			 gScratchItem;


CBitMap			*gScratchPad;			/* Image window editing				*/


CursHandle		gSelectCurs;
CursHandle		gDragCurs;
CursHandle		gPencilCurs;
CursHandle		gEraserCurs;



#define		kExtraMasters		6
#define		kRainyDayFund		90000
#define		kCriticalBalance	70000
#define		kToolboxBalance		20000




/***
 * ISymEditApp
 *
 *	Initialize the application.
 *
 ***/

void CSymEditApp::ISymEditApp(void)

{	
	CApplication::IApplication( kExtraMasters, kRainyDayFund, 
								kCriticalBalance, kToolboxBalance);
								
						
/*  The parameters to IApplication are the number of times to call  
    MoreMasters, the total number of bytes of heap space to reserve for                       
    monitoring low memory situations, and the portion of the memory
    reserve to set aside for critical operations and toolbox calls.
    
    Four (4) is a reasonable number of MoreMasters calls,                           
    but you should determine a good number for your application                     
    by observing the heap using Lightsbug,                                              
    TMON, or Macsbug. Set this parameter to zero, give your                         
    program a rigorous work-out, then look at the heap and count                        
    how many master pointer blocks have been allocated. Master                      
    pointer blocks are nonrelocatable and have a size of $100                           
    (hex). You should call MoreMasters at least this many                               
    times -- add a few extra just to be safe. The purpose of all                        
    this preflighting is to prevent heap fragmentation. You                             
    don't want the Memory Manager to call MoreMasters and                           
    create a nonrelocatable block in the middle of your heap. By                        
    calling MoreMasters at the very beginning of the program,                           
    you ensure that these blocks are allocated in a group at the                        
    bottom of the heap. 
                                                                        
    The memory reserve is a safeguard for handling low memory                   
    conditions and is used by the GrowMemory method in                              
    CApplication (check there for more comments). In general,                           
    your program should never request a memory block greater                        
    than this reserve size without explicitly checking in                               
    advance whether there is enough free memory to satisfy the                      
    the request.
                                                                                    
 */ 								
	
	gClipSym 	 = NULL;
	gScratchSym  = NULL;

	gClipRec 	 = NULL;
	gScratchRec	 = NULL;
	
	gClipItem	 = SELnone;
	gScratchItem = SELnone;

	gScratchPad = new CBitMap;
	gScratchPad->IBitMap(IMAP_WIDTH, IMAP_HEIGHT, TRUE);
	
	gSelectCurs = GetCursor(CURSselect);
	gDragCurs   = GetCursor(CURSdrag);
	gPencilCurs = GetCursor(CURSpencil);
	gEraserCurs = GetCursor(CURSeraser);
	
	gFileWind 	= NULL;

}



/***
 * SetUpFileParameters
 *
 *	In this routine, you specify the kinds of files your
 *	application opens.
 *
 *
 ***/

void CSymEditApp::SetUpFileParameters(void)

{
	inherited::SetUpFileParameters();	/* Be sure to call the default method */

		/**
		 **	sfNumTypes is the number of file types
		 **	your application knows about.
		 **	sfFileTypes[] is an array of file types.
		 **	You can define up to 4 file types in
		 **	sfFileTypes[].
		 **
		 **/

	sfNumTypes = 1;
	sfFileTypes[0] = 'SWSY';

		/**
		 **	Although it's not an instance variable,
		 **	this method is a good place to set the
		 **	gSignature global variable. Set this global
		 **	to your application's signature. You'll use it
		 **	to create a file (see CFile::CreateNew()).
		 **
		 **/

	gSignature = 'SWSE';
}



/***
 * DoCommand
 *
 *
 ***/
 
void CSymEditApp::DoCommand(long theCommand)

{
	short result;
	
	
	switch (theCommand) {
	
		case cmdAbout:
		
					PositionDialog('ALRT', ALRTaboutApp);	
					result = Alert(ALRTaboutApp, NULL);	
					break;
				
				
		default:
					inherited::DoCommand(theCommand);
					break;
	}
}



/***
 *
 * UpdateMenus 
 *
 *   Perform menu management tasks
 *
***/

void CSymEditApp::UpdateMenus()                    
	 
{
	inherited::UpdateMenus();
	 	
 	gBartender->EnableCmd(cmdAbout);

	gBartender->EnableCmd(cmdNew);
	gBartender->EnableCmd(cmdOpen);
	gBartender->EnableCmd(cmdQuit);
}



/***
 * CreateDocument
 *
 *	The user chose New from the File menu.
 *
 ***/

void CSymEditApp::CreateDocument()

{
	CFileWindow	*theDocument = NULL;
	
	TRY
	{
		theDocument = new CFileWindow;		
		theDocument->IFileWindow(this, FALSE);
		theDocument->NewFile();
	}
	
	CATCH
	{
		 if (theDocument) theDocument->Dispose();
	}
	ENDTRY;
}



/***
 * OpenDocument
 *
 *	The user chose Open from the File menu.
 *
 *	The macSFReply is a good SFReply record that contains
 *	the name and vRefNum of the file the user chose to
 *	open.
 *
 ***/
 
void CSymEditApp::OpenDocument(SFReply *macSFReply)

{
	CFileWindow	*theDocument = NULL;
	
	TRY
	{
		theDocument = new CFileWindow;
		theDocument->IFileWindow(this, FALSE);
		theDocument->OpenFile(macSFReply);
	}
	
	CATCH
	{
		 if (theDocument) theDocument->Dispose();
	}
	ENDTRY;
}