/******************************************************************************
 CArtClassApp.c

							The ArtClass Application Class
			
	SUPERCLASS = CApplication
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
		written by: Gregory H. Dow

 ******************************************************************************/

#include <Global.h>
#include <Commands.h>
#include <Constants.h>
#include <CFWDesktop.h>
#include <CBartender.h>
#include <CBitMap.h>
#include <CPNTGFile.h>
#include <TBUtilities.h>
#include "ArtClassCmds.h"
#include "CArtClassDoc.h"
#include "CArtClassApp.h"
#include "CToolsDir.h"
#include "CPatternsDir.h"
#include "CAboutBox.h"
#include "CAbstractText.h"
#include <Packages.h>

/*** Global Variables ***/

extern CDesktop		*gDesktop;			/* The visible Desktop				*/
extern CBartender	*gBartender;		/* Manages all menus				*/
extern OSType		gSignature;			/* Creator for Application's files	*/
extern long			gSleepTime;			/* Max time between events			*/

/*** Global Varialbes for Art Class ***/

CBitMap			*gScratchPad;			/* Offscreen drawing port & bitmap	*/
short			gPaintTool;				/* The current painting tool		*/
Pattern			gPattern;				/* The current fill pattern			*/
short			gLineSize;				/* The current line size			*/

short			gFontNum;				/* Font number						*/
short			gFontItem;				/* Item number in Font menu			*/
short			gFontSize;				/* Font size						*/
short			gSizeItem;				/* Item number in Size menu			*/
Style			gFontStyle;				/* Current font style				*/
long			gLineSpacing;			/* Line spacing cmd number			*/

CursHandle		gShapeCurs;
CursHandle		gPencilCurs;
CursHandle		gEraserCurs;
CursHandle		gBrushCurs;
CursHandle		gDragCurs;
CursHandle		gSelectCurs;

#define		CURSshape		200
#define		CURSpencil		201
#define		CURSeraser		202
#define		CURSbrush		203
#define		CURSdrag		204
#define		CURSselect		205

#define		kExtraMasters		4
#define		kRainyDayFund		60000
#define		kCriticalBalance	40000
#define		kToolboxBalance		20000

/******************************************************************************
 IArtClassApp

		Initialize the ArtClass Application
 ******************************************************************************/

void	CArtClassApp::IArtClassApp()
{
	Str255			itemStr;

	gPaintTool = toolPENCIL;
										/* Initialize superclass			*/
	CApplication::IApplication( kExtraMasters, kRainyDayFund, 
					kCriticalBalance, kToolboxBalance);

										/* Global variable for an offscreen	*/
										/*   BitMap with its own port		*/
	gScratchPad = new(CBitMap);
	gScratchPad->IBitMap(PNTG_WIDTH, PNTG_HEIGHT, TRUE);
	
	itsPatsDir->GetPattern(&gPattern);
	gLineSize = 1;
	
	gShapeCurs = GetCursor(CURSshape);
	gPencilCurs = GetCursor(CURSpencil);
	gEraserCurs = GetCursor(CURSeraser);
	gBrushCurs = GetCursor(CURSbrush);
	gDragCurs = GetCursor(CURSdrag);
	gSelectCurs = GetCursor(CURSselect);
	
	gFontNum = applFont;
	gFontSize = SYS_FONT_SIZE;
	gFontStyle = NOTHING;
	gLineSpacing = cmdSingleSpace;
	
	GetFontName(gFontNum, itemStr);
	gFontItem = gBartender->FindItemText(MENUfont, itemStr);
	
	NumToString((long) gFontSize, itemStr);
	gSizeItem = gBartender->FindItemText(MENUsize, itemStr);
	
		// The task names used in the undo menu item are stored
		// in STR# 130, in whatever order you define. If you want
		// text editing to display the correct name, e.g. "Undo Typing"
		// then you must tell CAbstractText where its strings are in
		// the STR# 130 resource.
		
	CAbstractText::cFirstTaskIndex = 1;	// first in the resource

}


/******************************************************************************
 MakeDesktop {OVERRIDE}

		Use the Desktop subclass which supports floating windows
 ******************************************************************************/

void	CArtClassApp::MakeDesktop()
{
	gDesktop = new(CFWDesktop);
	((CFWDesktop*)gDesktop)->IFWDesktop(this);
}
	

/******************************************************************************
 SetUpFileParameters {OVERRIDE}

		Set parameters used by the Standard File Package
 ******************************************************************************/

void	CArtClassApp::SetUpFileParameters()
{
	inherited::SetUpFileParameters();	/* Use defaults from superclass		*/
	
	gSignature = 'Art';				/* Signature for our application	*/

	sfNumTypes = 1;
	sfFileTypes[0] = 'PNTG';			/* ArtClass can read paintings		*/
}
	

/******************************************************************************
 SetUpMenus {OVERRIDE}

		Set up the menus used by the Buffet Application
 ******************************************************************************/

void	CArtClassApp::SetUpMenus()
{
	MenuHandle		macMenu;
	
	itsToolsDir = new(CToolsDir);		/* Director for a tear-off tool		*/
	itsToolsDir->IToolsDir(this);		/*   menu							*/

	itsPatsDir = new(CPatternsDir);		/* Tear-off Patterns menu			*/
	itsPatsDir->IPatternsDir(this);	

	inherited::SetUpMenus();
	
	AddResMenu(GetMHandle(MENUfont), 'FONT');
	
										/* No individual items are ever		*/
										/*   dimmed in these menus. It's	*/
										/*   all or nothing					*/
	gBartender->SetDimOption(MENUtools, dimNONE);
	gBartender->SetDimOption(MENUpatterns, dimNONE);
	gBartender->SetDimOption(MENUlines, dimNONE);
	gBartender->SetDimOption(MENUtext, dimNONE);
	gBartender->SetDimOption(MENUfont, dimNONE);
	gBartender->SetDimOption(MENUsize, dimNONE);
	
										/* These menus have check marked	*/
										/*   items							*/
	gBartender->SetUnchecking(MENUlines, TRUE);
	gBartender->SetUnchecking(MENUtext, TRUE);
	gBartender->SetUnchecking(MENUfont, TRUE);
	gBartender->SetUnchecking(MENUsize, TRUE);
}
	

/******************************************************************************
 DoCommand {OVERRIDE}

		Execute a command
 ******************************************************************************/

void	CArtClassApp::DoCommand(
	long		theCommand)
{
	CAboutBox	*theAboutBox;
	Str63		itemName;
	long		number;
	Style		aStyle;
	
	if (theCommand < 0) {		
		switch (HiShort(-theCommand)) {
		
			case MENUtools:
				gPaintTool = LoShort(-theCommand);
				itsToolsDir->SelectTool(gPaintTool);
				break;
				
			case MENUpatterns:
				itsPatsDir->SelectPattern(LoShort(-theCommand));
				itsPatsDir->GetPattern(&gPattern);
				break;
		
			case MENUfont:
				gBartender->GetCmdText(theCommand, itemName);
				gFontItem = LoShort(-theCommand);
				GetFontNumber(itemName, &gFontNum);
				break;
				
			case MENUsize:
				gBartender->GetCmdText(theCommand, itemName);
				StringToNum(itemName, &number);
				gSizeItem = LoShort(-theCommand);
				gFontSize = number;
				break;

			default:
				inherited::DoCommand(theCommand);
		}

	} else {
		switch (theCommand) {
			
			case cmdAbout:
				theAboutBox = new(CAboutBox);
				theAboutBox->IAboutBox(this);
				theAboutBox->Dispose();
				FlushEvents(mDownMask+mUpMask+keyDownMask+keyUpMask
								+autoKeyMask, 0);
				break;

			case cmdLineNone:
			case cmdLineOne:
			case cmdLineTwo:
			case cmdLineThree:
			case cmdLineFour:
				gLineSize = theCommand - cmdLineNone;
				break;
				
			case cmdPlain:
				gFontStyle = NOTHING;
				break;
				
			case cmdBold:					/* Styles are set by flipping bits	*/
			case cmdItalic:					/*   Bold is bit 0, Italic is bit 1	*/
			case cmdUnderline:				/*   Underline is bit 2, etc.		*/
			case cmdOutline:				/*   By cleverly numbering these	*/
			case cmdShadow:					/*   commands, we can just shift	*/
			case cmdCondense:				/*   a bit to get the proper style	*/
			case cmdExtend:
				aStyle = 1 << (theCommand - cmdBold);
					gFontStyle ^= aStyle;	/* Toggle style characteristic		*/
											/*   by XOR'ing proper bit			*/
				break;
				
			case cmdSingleSpace:
			case cmd1HalfSpace:
			case cmdDoubleSpace:
				gLineSpacing = theCommand;
				break;
	
			default:
				inherited::DoCommand(theCommand);
				break;
		}
	}
}
	

/******************************************************************************
 UpdateMenus {OVERRIDE}

		Perform menu management tasks
 ******************************************************************************/

void	CArtClassApp::UpdateMenus()
{
	inherited::UpdateMenus();

	gBartender->CheckMarkCmd(gLineSize + cmdLineNone, TRUE);
}
	

/******************************************************************************
 CreateDocument {OVERRIDE}
 
		Make a document in response to the "New" menu selection.
 ******************************************************************************/

void	CArtClassApp::CreateDocument()
{
	CArtClassDoc	*theDocument = NULL;
	
	TRY
	{
		theDocument = new(CArtClassDoc);
		theDocument->IArtClassDoc(this);
		theDocument->NewFile();
	}
	CATCH
	{
		ForgetObject( theDocument);
	}
	ENDTRY;
}	


/******************************************************************************
 OpenDocument {OVERRIDE}
 
		Open an existing file and create a document object for displaying
		information.
 ******************************************************************************/

void	CArtClassApp::OpenDocument(
	SFReply		*macSFReply)
{
	CArtClassDoc	*theDocument = NULL;
	
	TRY
	{
		theDocument = new(CArtClassDoc);
		theDocument->IArtClassDoc(this);
		theDocument->OpenFile(macSFReply);
	}
	CATCH
	{
		ForgetObject( theDocument);
	}
	ENDTRY;
}


/******************************************************************************
 SwitchToDA
 
		A DA is becoming active
 ******************************************************************************/

void	CArtClassApp::SwitchToDA()
{
	inherited::SwitchToDA();
	
	gBartender->DisableMenu(MENUtools);
	gBartender->DisableMenu(MENUpatterns);
	gBartender->DisableMenu(MENUlines);
	gBartender->DisableMenu(MENUtext);
}


/******************************************************************************
 SwitchFromDA
 
		Our application is becoming active after a DA had been running
 ******************************************************************************/

void	CArtClassApp::SwitchFromDA()
{
	inherited::SwitchFromDA();
	
	gBartender->EnableMenu(MENUtools);
	gBartender->EnableMenu(MENUpatterns);
	gBartender->EnableMenu(MENUlines);
	gBartender->EnableMenu(MENUtext);
}