
/*
 *  time.c
 *
 *  Copyright (c) 1991 Symantec Corporation.  All rights reserved.
 *
 */

#include <MacHeaders>

#include "time.h"
#include "stdio.h"
#include "string.h"

static void copy_in(struct tm *);
static void copy_out(struct tm *);
static short modulo(short *, short);

static DateTimeRec dt;
static short ndays[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
static short cumul[] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };

static char *dayname[] = {
	"Sunday", "Monday", "Tuesday", "Wednesday",
	"Thursday", "Friday", "Saturday"
};

static char *monthname[] = {
	"January", "February", "March", "April", "May", "June",
	"July", "August", "September", "October", "November", "December"
};


/* ---------- entry points ---------- */


/*
 *  clock - return tick count
 *
 */

clock_t
clock(void)
{
	return(TickCount());
}


/*
 *  difftime - compare two times
 *
 */

double
difftime(time_t time1, time_t time0)
{
	if (time0 > time1)
		return(- (double) (time0 - time1));
	return(time1 - time0);
}


/*
 *  mktime - validate/convert broken-down time
 *
 */

time_t
mktime(struct tm *t)
{
	time_t when;

	copy_in(t);
	if (dt.year < 1904 || dt.year > 2040)
		return(-1);
	asm {
		lea		dt,a0
		_Date2Secs
		move.l	d0,when
	}
	copy_out(t);
	return(when);
}


/*
 *  time - tell the time
 *
 */

time_t
time(time_t *timer)
{
	time_t now = Time;
	
	if (timer)
		*timer = now;
	return(now);
}


/*
 *  asctime - construct standard time string
 *
 */

char *
asctime(const struct tm *t)
{
	static char buf[26];

	strftime(buf, sizeof buf, "%c\n", t);
	return(buf);
}


/*
 *  ctime - construct standard time string
 *
 */

char *
ctime(const time_t *timer)
{
	return(asctime(localtime(timer)));
}


/*
 *  gmtime - convert to GMT
 *
 */

struct tm *
gmtime(const time_t *timer)
{
	return(NULL);		/*  no can do  */
}


/*
 *  localtime - convert to local time
 *
 */

struct tm *
localtime(const time_t *timer)
{
	static struct tm tm;
	
	Secs2Date(*timer, &dt);
	copy_out(&tm);
	return(&tm);
}


/*
 *  strftime - construct custom time string
 *
 */

size_t
strftime(char *target, size_t max, const char *fmt, register const struct tm *t)
{
	register size_t nwritten = 0, n;
	register int c, i;
	register const char *s;
	char buf[26];
	
	for (c = *fmt; c; c = *++fmt) {
		if (c != '%')
			goto copy1;
		s = buf;
		n = 0;
		switch (c = *++fmt) {
			case 'a':
				s = dayname[t->tm_wday];
				n = 3;
				break;
			case 'A':
				s = dayname[t->tm_wday];
				break;
			case 'b':
				s = monthname[t->tm_mon];
				n = 3;
				break;
			case 'B':
				s = monthname[t->tm_mon];
				break;
			case 'c':
				strftime(buf, sizeof buf, "%a %b %D %H:%M:%S %Y", t);
				break;
			case 'd':
				i = t->tm_mday;
				goto digits2;
			case 'D':
				sprintf(buf, "%2d", t->tm_mday);
				break;
			case 'H':
				i = t->tm_hour;
				goto digits2;
			case 'I':
				i = ((t->tm_hour + 11) % 12) + 1;
				goto digits2;
			case 'j':
				sprintf(buf, "%.3d", t->tm_yday + 1);
				break;
			case 'm':
				i = t->tm_mon + 1;
				goto digits2;
			case 'M':
				i = t->tm_min;
				goto digits2;
			case 'p':
				s = t->tm_hour < 12 ? "AM" : "PM";
				break;
			case 'S':
				i = t->tm_sec;
				goto digits2;
			case 'U':
				i = (t->tm_yday - t->tm_wday + 7) / 7;
				goto digits2;
			case 'w':
				sprintf(buf, "%d", t->tm_wday);
				break;
			case 'W':
				i = (t->tm_yday - ((t->tm_wday + 6) % 7) + 7) / 7;
				goto digits2;
			case 'x':
				sprintf(buf, "%d/%d/%.2d", t->tm_mon + 1, t->tm_mday,
					t->tm_year % 100);
				break;
			case 'X':
				sprintf(buf, "%d:%.2d %cM", ((t->tm_hour + 11) % 12) + 1,
					t->tm_min, "AP"[t->tm_hour / 12]);
				break;
			case 'y':
				i = t->tm_year % 100;
				/* ... */
			digits2:
				sprintf(buf, "%.2d", i);
				n = 2;
				break;
			case 'Y':
				sprintf(buf, "%d", 1900 + t->tm_year);
				break;
			case 'Z':
				continue;
			default:
				if (c != '%')
					return(0);
			copy1:
				s = fmt;
				n = 1;
				break;
		}
		if (n == 0)
			n = strlen(s);
		if ((nwritten += n) >= max)
			return(0);
		sprintf(target, "%.*s", (int) n, s);
		target += n;
	}
	return(nwritten);
}


/* ---------- utility routines ---------- */


/*
 *  copy_in - convert "struct tm" to "DateTimeRec"
 *
 */

static void
copy_in(register struct tm *t)
{
	register int n;
	
		/*  force seconds, minutes, hours, month into range  */
	
	dt.second = t->tm_sec;
	dt.minute = t->tm_min + modulo(&dt.second, 60);
	dt.hour = t->tm_hour + modulo(&dt.minute, 60);
	dt.day = t->tm_mday + modulo(&dt.hour, 24);
	dt.month = t->tm_mon;
	dt.year = t->tm_year + modulo(&dt.month, 12);
	
		/*  force day-of-month into range (too high)  */
		
	if (dt.day > 0) {
		for (;;) {
			n = ndays[dt.month];
			if (dt.month == 1 && !(dt.year & 3))
				++n;
			if (dt.day <= n)
				break;
			dt.day -= n;
			if (++dt.month == 12) {
				dt.month = 0;
				++dt.year;
			}
		}
	}
	
		/*  force day-of-month into range (too low)  */
		
	else {
		do {
			if (--dt.month < 0) {
				dt.month = 11;
				--dt.year;
			}
			n = ndays[dt.month];
			if (dt.month == 1 && !(dt.year & 3))
				++n;
		} while ((dt.day += n) <= 0);
	}
	
		/*  offset year/month  */
	
	dt.year += 1900;
	++dt.month;
}


/*
 *  copy_out - convert "DateTimeRec" to "struct tm"
 *
 */

static void
copy_out(register struct tm *t)
{
		/*  copy/offset fields  */

	t->tm_sec = dt.second;
	t->tm_min = dt.minute;
	t->tm_hour = dt.hour;
	t->tm_mday = dt.day;
	t->tm_mon = dt.month - 1;
	t->tm_year = dt.year - 1900;
	t->tm_isdst = -1;
		
		/*  compute day-of-year  */
		
	t->tm_yday = cumul[t->tm_mon] + t->tm_mday - 1;
	if (t->tm_mon > 1 && !(t->tm_year & 3))
		++t->tm_yday;
		
		/*  compute day-of-week  */
		
	t->tm_wday = ((((t->tm_year - 1) * 5) >> 2) + t->tm_yday + 2) % 7;
}


/*
 *  modulo - compute quotient/remainder
 *
 *  The quotient is returned.  The remainder replaces the dividend.
 *
 *  The remainder is guaranteed to be non-negative (assuming the divisor
 *  is positive).
 *
 */

static short
modulo(short *np, short d)
{
	asm {
		movea.l	np,a1
		move.w	(a1),d1
		ext.l	d1
		divs	d,d1
		move.w	d1,d0
		swap	d1
		tst.w	d1
		bpl.s	@1
		add.w	d,d1
		subq.w	#1,d0
@1		move.w	d1,(a1)
	}
}
