/******************************************************************************
 CAppleEvent.c
	
	A class that encapsulates an incoming AppleEvent and its default reply.
	AppleEvents are received by the application's switchboard and packaged
	into a CAppleEvent object by the application. The CAppleEvent object
	is then passed down the chain of command, just like menu commands and
	keystrokes. CApplication handles the four required AppleEvents. To handle
	other AppleEvents, override the DoAppleEvent method in your pane, director,
	document, or application subclass.
		
		Copyright  1991 Symantec Corporation. All rights reserved.
		
 ******************************************************************************/


#include "CAppleEvent.h"
#include "CArray.h"
#include "CSwitchboard.h"
#include "TCLUtilities.h"

/******************************************************************************
 IAppleEvent
 
 	Initialize a CAppleEvent object. theEvent is the AppleEvent as received by 
 	CSwitchboard. theReply is the default reply the AppleEvent manager generates
 	when an AppleEvent is sent. theRefCon is the reference value associated
 	with the particular event class and ID. Normally in the TCL it is zero.
 	eventClass and eventID are simply the event class and ID which were extracted
 	from theEvent.
 	
******************************************************************************/

void CAppleEvent::IAppleEvent( const AppleEvent *theEvent, const AppleEvent *theReply,
					long theRefCon, DescType eventClass, DescType eventID)
{
	
	this->theEvent = *theEvent;
	this->theReply = *theReply;
	this->theRefCon = theRefCon;
	this->eventClass = eventClass;
	this->eventID = eventID;
	
	notificationRec = NULL;
	idleProc = NULL;
	
		// canInteract indicates whether user interaction has been
		// requested and approved.
		
	canInteract = FALSE;
	
		// errCode is used to store the error code ultimately
		// returned back to the AppleEvent mgr, which in turn will
		// pass it to the sender. It is initialized to errAEEventNotHandled
		// to indicate that the event has not yet been handled. If no object in
		// the chain of command handles an event, then errCode will
		// not be changed, and the sender will be informed that the
		// event wasn't handled.
		
	errCode = errAEEventNotHandled;
	
}	/* CAppleEvent::IAppleEvent */

/******************************************************************************
 GetAEEvent
 
 	Return the AppleEvent
******************************************************************************/

AppleEvent CAppleEvent::GetAEEvent( void)
{
	return theEvent;
	
}	/* CAppleEvent::GetAEEvent */

/******************************************************************************
 GetAEReply
 
 	Return the AppleEvent reply
******************************************************************************/

AppleEvent CAppleEvent::GetAEReply( void)
{
	return theReply;
	
}	/* CAppleEvent::GetAEReply */

/******************************************************************************
 GetAERefCon
 
 	Return the AppleEvent reference value, which is associated with a given
 	event class and ID when AEInstallEventHandler is called. The TCL always
 	initializes this to zero, but other values could be used if your
 	application calls AEInstallHandler itself.
 	
******************************************************************************/

long CAppleEvent::GetAERefCon( void)
{
	return theRefCon;
	
}	/* CAppleEvent::GetAERefCon */

/******************************************************************************
 GetEventClass
 
 	Return the event class, e.g. 'aevt'. The event class and ID together
 	uniquely determine a particular AppleEvent.
******************************************************************************/

DescType CAppleEvent::GetEventClass( void)
{
	return eventClass;
}

/******************************************************************************
 GetEventID
 
 	Return the event ID, e.g. 'oapp'.
******************************************************************************/

DescType CAppleEvent::GetEventID( void)
{
	return eventID;
}

/******************************************************************************
 GetDescList
 
 	Return the AEDescList associated with the keyword whichParam.
******************************************************************************/

void CAppleEvent::GetDescList( AEKeyword whichParam, AEDescList *descList)
{
	AEDescList	theList;
	AppleEvent	event = theEvent;
	
	FailOSErr( AEGetParamDesc( &event, whichParam, typeAEList, &theList));
	
	*descList = theList;
	
}	/* CAppleEvent::GetDescList */

/******************************************************************************
 ExtractFromDescList
 
 	Gets the descriptor list associated with the parameter whichParam,
 	and extracts all the items in the list into a CArray. This is a
 	useful utility if all the items in the list are the same size
 	and the data in the list is not keyword separated - the keywords
 	are not used here.
******************************************************************************/

CArray *CAppleEvent::ExtractFromDescList( AEKeyword whichParam, DescType itemType,
					Size itemSize)
{
	AEDescList	descList;
	CArray		*array = NULL;
	long		itemCount, i;
	AEKeyword	keyword;
	DescType	actualType;
	Size		actualSize;
	Handle		bufH = NULL;
	Ptr			itemBuffer;
	
	TRY
	{
		descList.dataHandle = NULL;
		GetDescList( whichParam, &descList);
		array = new( CArray);
		array->IArray( itemSize);
		
		bufH = NewHandleCanFail( itemSize);
		FailNIL( bufH);
		
		MoveHHi( bufH);
		HLock( bufH);
		itemBuffer = *bufH;
		
		FailOSErr( AECountItems( &descList, &itemCount));
		for( i = 1; i <= itemCount; i++)
		{
			FailOSErr( AEGetNthPtr( &descList, i, itemType, &keyword, &actualType,
								itemBuffer, itemSize, &actualSize));
								
			array->InsertAtIndex( itemBuffer, i);
		}
	}
	CATCH
	{
		ForgetHandle( bufH);
		ForgetObject( array);
		AEDisposeDesc( &descList);
	}
	ENDTRY;
	
	ForgetHandle( bufH);
	AEDisposeDesc( &descList);
	return array;

}	/* CAppleEvent::ExtractFromDescList */

/******************************************************************************
 GetErrorResult
 	
 	Return the error code for this event. It will be noErr if the
 	event was handled successfully, errAEEventNotHandled if the event
 	was not handled, or some other result code if someone attempted to handle 
    it, failed, and stored a result code. Note that CSwitchboard uses a special
    exception handler to trap exceptions raised during the handling of an
    AppleEvent, and will return the exception error instead.
    
******************************************************************************/

OSErr CAppleEvent::GetErrorResult( void)
{
	return errCode;
}

/******************************************************************************
 SetErrorResult
 
 	Set the result code for this event
******************************************************************************/

void CAppleEvent::SetErrorResult( OSErr anErrorCode)
{
	errCode = anErrorCode;
	
}	/* CAppleEvent::SetErrorResult */

/******************************************************************************
 GotRequiredParams
 
 	Does the recommended check to determine if all the required parameters
 	of an AppleEvent have already been extracted. You should call this
 	method after you extract all the parameters you know about, and before
 	you actually handle the event.
 	
******************************************************************************/

Boolean CAppleEvent::GotRequiredParams( void)
{
   DescType returnedType;
   Size 	actualSize;
   OSErr	err;

   err = AEGetAttributePtr ( &theEvent, keyMissedKeywordAttr, 
						typeWildCard, &returnedType, NULL, 0, 
						&actualSize);
   
   return err == errAEDescNotFound;
   
 }	/* CAppleEvent::GotRequiredParams */
 
 /******************************************************************************
 RequestInteraction
 
 	Indicates to the OS that you need to interact with the user (e.g. put
 	up a dialog) in order to handle the event. If interaction is allowed,
 	and if the application is brought forward before the timeout period
 	has elapsed, then noErr is returned and you can proceed with the interaction.
 	If any other value is returned, then interaction is not allowed or your
 	timeout period elapsed (check the result code).
 	
 	CSwitchboard provides a default AppleEvent idle proc that can be used
 	with AEInteractWithUser or AESend. To provide a custom one, put a pointer
 	to it in the idleproc instance variable. The AppleEvent mgr provides
 	a default, notification mgr record if NULL is passed to AEInteractWithUser.
 	The TCL uses the default, but you can provide a custom one by providing a
 	pointer to it in the notificationRec instance variable.
 	
******************************************************************************/

OSErr CAppleEvent::RequestInteraction( long timeOutTicks)
{
	OSErr		err;
	IdleProcPtr proc;
	
		// Don't call AEInteractWithUser again if it has already been
		// requested and approved.
		
	if (!canInteract)
	{
		if (this->idleProc)
			proc = this->idleProc;
		else
			proc = (IdleProcPtr)AppleEventIdleProc;
			
		err = AEInteractWithUser( timeOutTicks, (NMRecPtr) notificationRec, proc);
		if (err == noErr)
			canInteract = TRUE;
	}
	else
		err = noErr;
	return err;

}	/* CAppleEvent::RequestInteraction */
