/******************************************************************************
 CBartender.c

							The Bartender Class
							
		Manages menus and the menu bar. Maintains a table which maps menu
		choices into command numbers.
	
	SUPERCLASS = CObject
	
	Copyright  1989 Symantec Corporation. All rights reserved.

	TCL 1.1 CHANGES
	[
		- implemented EnableMenuBar and DisableMenuBar methods
	]
	
 ******************************************************************************/

#include "Global.h"
#include "Constants.h"
#include "CBartender.h"
#include "Commands.h"
#include "CBureaucrat.h"
#include "CMBarChore.h"
#include "CApplication.h"
#include "TCLUtilities.h"

#include <Packages.h>

/*** Global Variables ***/

extern CBureaucrat	*gGopher;			/* First in line to get commands	*/
extern CApplication *gApplication;

/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/
	

/******************************************************************************
 IBartender

		Initialize a Bartender object
 ******************************************************************************/

void	CBartender::IBartender(
	short		MBARid)					/* 'MBAR' Resource ID				*/
{
	short				**theMenuIDs;	/* List of MENU id numbers			*/
	short				MBARlength;		/* Number of menus in MBAR resource	*/
	register short		i;				/* Index into number of menus		*/
	short				MENUid;			/* MENU resource ID					*/
	MenuEntry			theEntry;		/* Entry in the menu table			*/
	
										/* Menus to install at startup are	*/
										/*   specified in a MBAR resource	*/
	theMenuIDs = (short**) GetResource('MBAR', MBARid);
	
	FailNILRes( theMenuIDs);
	MBARlength = (*theMenuIDs)[0];		/* MBAR resource is just an array	*/
										/*   of MENU id's (short's) where	*/
										/*   the first short is a count of	*/
										/*   the number of menus			*/
	
										/* Allocate a menu table with an	*/
										/*   entry for each menu			*/
	numMenus = MBARlength;
	theMenus = (MenuEntryH) NewHandleClear(numMenus * sizeof(MenuEntry));

										/* It's important to use a local	*/
										/*   variable to store the number	*/
										/*   of menus in the MBAR since the	*/
										/*   numMenus instance variable may	*/
										/*   change within the loop because	*/
										/*   of hierarchical menus			*/
										
	for (i = 0; i < MBARlength; i++) {	/* Run thru all the menus in MBAR	*/
		MENUid = (*theMenuIDs)[i+1];
		theEntry.MENUid = MENUid;
		theEntry.macMenu = (MenuHandle) GetMenu(MENUid);
		FailNILRes( theEntry.macMenu);
		theEntry.dimming = DEFAULT_DIM;
		theEntry.unchecking = FALSE;
		theEntry.hasHMenus = FALSE;
		theEntry.inMenuBar = TRUE;		/* mark menu as installed in menu bar. */
		theEntry.lastEnable = MenuEnabled( theEntry.macMenu);
		
										/* We append command numbers to the	*/
										/*   text of item names				*/
		ExtractCommands(&theEntry);
		(*theMenus)[i] = theEntry;
										/* Automatically install any hier	*/
										/*   menus							*/
		ExtractHierMenus(theEntry.macMenu, i);
										/* Build up the menu bar from left	*/
										/*   to right						*/
		InsertMenu(theEntry.macMenu, 0);
	}
	
	ReleaseResource( (Handle) theMenuIDs);		/* Don't need MBAR resource anymore	*/
	
	DrawMenuBar();						/* Draw titles of the menus we		*/
										/*   inserted						*/
	
	choreAssigned = FALSE;
	forceMBarUpdate = FALSE;

}


/**** I N S E R T I O N / D E L E T I O N   M E T H O D S ****/


/******************************************************************************
 AddMenu

		Add a menu to the menu table, if its not already in it, and optionally
		install the menu in the menu bar.
 ******************************************************************************/

void	CBartender::AddMenu(
	short		MENUid,					/* MENU Resource ID					*/
	Boolean		install,				/* Install menu in menu bar?		*/
	short		beforeID)				/* MENU ID of menu to put this one	*/
										/*   in front of					*/
{
	register short		i;				/* Index into menu table			*/
	MenuEntry			theEntry;		/* Entry in the menu table			*/
	

	i = FindMenuIndex(MENUid);			/* Check if menu is in our table	*/
	
	if (i < 0) {						/* Menu not in our table -- add it	*/

		i = numMenus++;					/* Save original value, then		*/
										/*   increment count of menus		*/
										/* Increase size of menu table		*/
		SetHandleSize((Handle) theMenus, numMenus * sizeof(MenuEntry));

										/* Fill in fields of a menu entry	*/
		theEntry.MENUid = MENUid;
		theEntry.macMenu = GetMenu(MENUid);
		FailNILRes( theEntry.macMenu);
		theEntry.dimming = DEFAULT_DIM;
		theEntry.unchecking = FALSE;
		theEntry.hasHMenus = FALSE;
		theEntry.lastEnable = MenuEnabled( theEntry.macMenu);
		ExtractCommands(&theEntry);
		(*theMenus)[i] = theEntry;
		ExtractHierMenus(theEntry.macMenu, i);

	}
	
	if (install) {						/* Install menu in the menu bar		*/
		(*theMenus)[i].inMenuBar = TRUE;	/* mark menu as installed */
		InsertMenu((*theMenus)[i].macMenu, beforeID);
		if (beforeID != hierMenu) {		/* Don't redraw if its a hier menu	*/
			forceMBarUpdate = TRUE;
			DeleteFromBar(NOTHING);		/* Force redraw of menu bar			*/
		}
	}
}
	

/******************************************************************************
 RemoveMenu

		Remove a menu from the menu bar and the menu table. The associated
		Toolbox menu record is also disposed.
 ******************************************************************************/

void	CBartender::RemoveMenu(
	short		MENUid)					/* MENU Resource ID					*/
{
	register short		i;				/* Index into menu table			*/
	
	DeleteFromBar(MENUid);				/* Remove from menu bar				*/
	
	i = FindMenuIndex(MENUid);			/* Locate menu in the menu table	*/
	if (i >= 0) {
										/* Free Toolbox menu record			*/
		ReleaseResource((Handle) (*theMenus)[i].macMenu);
										/* release commands handle			*/
		ForgetHandle( (*theMenus)[i].theCommands);
		
		numMenus--;						/* Reduce count of menus by one		*/
		if (i < numMenus) {
										/* If it's not the last entry,		*/
										/*   push all entries above it down	*/
										/*   one notch						*/
			BlockMove(*theMenus+i+1, *theMenus+i,
							(long) (numMenus - i) * sizeof(MenuEntry));
		}
										/* Reduce size of menu table		*/
		SetHandleSize((Handle) theMenus, numMenus * sizeof(MenuEntry));
	
	}	
										
}
	

/******************************************************************************
 InsertInBar

		Place a menu in the menu bar. Same as the AddMenu method with
		install set to TRUE.
 ******************************************************************************/

void	CBartender::InsertInBar(
	short		MENUid,					/* MENU Resource ID					*/
	short		beforeID)				/* MENU ID of menu to put this one	*/
										/*   in front of					*/
{
	AddMenu(MENUid, TRUE, beforeID);
}
	

/******************************************************************************
 DeleteFromBar

		Remove a menu from the menu bar, but leave it in the menu table.
 ******************************************************************************/

void	CBartender::DeleteFromBar(
	short		MENUid)					/* MENU Resource ID					*/
{
	short	menuIndex;
	CMBarChore *theChore;
	Boolean wasInBar = TRUE;
	
	if (MENUid != NOTHING) {
		DeleteMenu(MENUid);				/* Remove menu from the menu bar	*/
		forceMBarUpdate = TRUE;			/* must redraw menu bar				*/
		menuIndex = FindMenuIndex( MENUid); /* find menu in table */
		if (menuIndex >= 0)
		{
			wasInBar = (*theMenus)[menuIndex].inMenuBar;
			(*theMenus)[menuIndex].inMenuBar = FALSE; /* mark menu as not in menu bar */
		}
	}
	
	if (wasInBar && !choreAssigned) {
		choreAssigned = TRUE;
		theChore = new(CMBarChore);
		gApplication->AssignUrgentChore(theChore);
	}
}
	

/******************************************************************************
 InsertHierMenu

		Attach a hierarchical menu withing an existing menu. The command
		number for the hier menu is needed if you want to be able to
		dim/highlight the entire hier menu. The title of the hier menu
		is used as the text for the menu item within its parent menu.
 ******************************************************************************/

void	CBartender::InsertHierMenu(
	short		hMENUid,				/* MENU Resource ID of hier menu	*/
	long		cmdNo,					/* Command number for hier menu		*/
	short		inMENUid,				/* Menu in which to add hier menu	*/
	short		afterItem)				/* Item after which to insert hmenu	*/
{
    
    MenuHandle      macMenu;
    short           i, inMenuNumCmds;  
	
	i = FindMenuIndex(inMENUid);
	
	if (i >= 0) {
		AddMenu(hMENUid, TRUE, hierMenu);
		macMenu = GetMHandle(hMENUid);
		
        inMenuNumCmds = (*theMenus)[i].numCmds;
        
        if (afterItem < 0)
            afterItem = 0;
        
        else if (afterItem > inMenuNumCmds)
            afterItem = inMenuNumCmds;
        
        InsertMenuCmd(cmdNo, (**macMenu).menuData, inMENUid, afterItem);
        
        macMenu = GetMHandle(inMENUid);
		SetItemMark(macMenu, afterItem+1, hMENUid);
		SetItemCmd(macMenu, afterItem+1, hMenuCmd);
	
		(*theMenus)[i].hasHMenus = TRUE;
	}
}


/**** I T E M   M A N I P U L A T I O N  M E T H O D S ****/


/******************************************************************************
 EnableCmd

		Enable the menu item corresponding to the specified command number
 ******************************************************************************/

void	CBartender::EnableCmd(
	long		cmdNo)
{
	short		MENUid;
	MenuHandle	macMenu;
	short		itemNo;

	FindMenuItem(cmdNo, &MENUid, &macMenu, &itemNo);
	if (macMenu != NULL) {
		EnableItem(macMenu, itemNo);
	}
}
	

/******************************************************************************
 DisableCmd

		Disable the menu item corresponding to the specified command number
 ******************************************************************************/

void	CBartender::DisableCmd(
	long		cmdNo)
{
	short		MENUid;
	MenuHandle	macMenu;
	short		itemNo;

	FindMenuItem(cmdNo, &MENUid, &macMenu, &itemNo);
	if (macMenu != NULL) {
		DisableItem(macMenu, itemNo);
	}
}
	

/******************************************************************************
 EnableMenu

		Enable a menu as a whole. Note that items disabled separately, such
		as by using the DisableCmd method, will remain disabled. Uninstalled
		menus cannot be enabled.
 ******************************************************************************/

void	CBartender::EnableMenu(
	short		MENUid)					/* MENU resource ID					*/
{
	short			i;
	MenuHandle		macMenu;
	
	i = FindMenuIndex( MENUid);
	if (i >= 0)
	{
		macMenu = (*theMenus)[i].macMenu;
		if (!MenuEnabled( macMenu))
		{								/* Make sure menu is installed		*/
										/* and not already enabled			*/
			EnableItem(macMenu, 0);		/* Item 0 means the menu as a whole	*/
			DeleteFromBar(NOTHING);		/* Force redraw of menu bar			*/
		}
	}	
}

/******************************************************************************
 DisableMenu

		Disable a menu as a whole. The menu title and all items in the
		menu will be dimmed. Uninstalled menus cannot be disabled.
 ******************************************************************************/

void	CBartender::DisableMenu(
	short		MENUid)					/* MENU resource ID					*/
{
	MenuHandle		macMenu;
	
	macMenu = GetMHandle(MENUid);		/* Get associated Toolbox record	*/
	
	if ((macMenu != NULL) && MenuEnabled( macMenu))
	{									/* Make sure menu is installed		*/
										/* and not already disabled			*/
		DisableItem(macMenu, 0);		/* Item 0 means the menu as a whole	*/
		DeleteFromBar(NOTHING);			/* Force redraw of menu bar			*/
	}
}
	
/******************************************************************************
 EnableMenuBar
 
 	Enables all the menus currently in the menu bar.
******************************************************************************/
void CBartender::EnableMenuBar( void)
{
	short	currMenu;
	
	for ( currMenu = 0; currMenu < numMenus; currMenu++)
	{
		if ((*theMenus)[currMenu].inMenuBar)
			EnableMenu( (*theMenus)[currMenu].MENUid);
	}
}

/******************************************************************************
 DisableMenuBar
 
 	Disables all the menus currently in the menu bar.
******************************************************************************/
void CBartender::DisableMenuBar( void)
{
	short	currMenu;
	
	for ( currMenu = 0; currMenu < numMenus; currMenu++)
	{
		if ((*theMenus)[currMenu].inMenuBar)
			DisableMenu( (*theMenus)[currMenu].MENUid);
	}
}

/******************************************************************************
 SetCmdText

		Specify the text for a menu command
 ******************************************************************************/

void	CBartender::SetCmdText(
	long		cmdNo,					/* Command number of menu item		*/
	Str255		theText)				/* New text for the menu item		*/
{
	short				MENUid;
	MenuHandle			macMenu;
	short				itemNo;
	
	FindMenuItem(cmdNo, &MENUid, &macMenu, &itemNo);
	if (macMenu != NULL)
		SetItem(macMenu, itemNo, theText);
}
	

/******************************************************************************
 GetCmdText

		Retrieve the text for a menu command
 ******************************************************************************/

void	CBartender::GetCmdText(
	long		cmdNo,					/* Command number of menu item		*/
	Str255		theText)				/* Text of the menu item			*/
{
	short				MENUid;
	MenuHandle			macMenu;
	short				itemNo;
	
	FindMenuItem(cmdNo, &MENUid, &macMenu, &itemNo);
	if (macMenu != NULL)
		GetItem(macMenu, itemNo, theText);
}


/******************************************************************************
 CheckMarkCmd

		Place or remove a check mark before the menu item corresponding
		to a specified command. If checked is TRUE, a check mark is
		placed; if checked is FALSE, the check mark is removed.
 ******************************************************************************/

void	CBartender::CheckMarkCmd(
	long		cmdNo,
	Boolean		checked)				/* Check or uncheck?				*/
{
	short				MENUid;
	MenuHandle			macMenu;
	short				itemNo;

	FindMenuItem(cmdNo, &MENUid, &macMenu, &itemNo);
	if (MENUid != NOTHING) {
		CheckItem(macMenu, itemNo, checked);
	}
}


/**** I T E M   I N S E R T I O N / D E L E T I O N   M E T H O D S ****/


/******************************************************************************
 InsertMenuCmd

		Insert a menu command in a menu at a position following a specified
		item
 ******************************************************************************/

void	CBartender::InsertMenuCmd(
	long		cmdNo,					/* Command number of new menu item	*/
	Str255		theText,				/* Text for the menu item			*/
	short		MENUid,					/* Menu in which to insert cmd		*/
	short		afterItem)				/* Put new item after this one		*/
{
	short				i;				/* Index into menu table			*/
	register MenuEntryP	theEntry;		/* Entry in the menu table			*/
	register long		*cmdList;		/* List of menu commands			*/
	
	i = FindMenuIndex(MENUid);			/* Locate Menu in the menu table	*/
	
	if (i >= 0) {						/* If Menu is in the menu table ...	*/
		HLock((Handle) theMenus);		/* Lock table and extract the 		*/
		theEntry = &(*theMenus)[i];		/*   entry of interest				*/
		
		if (afterItem < 0) {			/* afterItem must be between 0 and	*/
			afterItem = 0;				/*   the total number of items		*/
		} else if (afterItem > theEntry->numCmds) {
			afterItem = theEntry->numCmds;
		}

										/* Add item to Mac menu				*/
		InsMenuItem(theEntry->macMenu, theText, afterItem);
		
										/* Now add new command to our list	*/
										
		theEntry->numCmds++;			/* Make room for a new item in list	*/
                
        if (theEntry->numCmds == 1)
            theEntry->theCommands = (long**) NewHandle(4L);
        else
            SetHandleSize((Handle) theEntry->theCommands, theEntry->numCmds * 4L);
        
		cmdList = *(theEntry->theCommands);
		
										/* Shift items down one slot		*/
		if (afterItem < theEntry->numCmds - 1) {
			BlockMove(cmdList + afterItem, cmdList + afterItem + 1,
					  (theEntry->numCmds - afterItem - 1) * 4L);
		}
		
		cmdList[afterItem] = cmdNo;		/* Store command number. Note that	*/
										/*   location [afterItem] is really	*/
										/*   the afterItem+1 item in the 	*/
										/*   list (difference between zero	*/
										/*   and one-based lists			*/
		HUnlock((Handle) theMenus);
	}
}


/******************************************************************************
 RemoveMenuCmd

		Remove the specified command from the Bartender.
 ******************************************************************************/

void	CBartender::RemoveMenuCmd(
	long		cmdNo)					/* Cmd number of item to remove		*/
{
	short		MENUid;
	MenuHandle	macMenu;
	short		itemNo;
	short		i;
	register MenuEntryP	theEntry;		/* Entry in the menu table			*/
	register long		*cmdList;		/* List of menu commands			*/

										/* Look up command in menu table	*/
	FindMenuItem(cmdNo, &MENUid, &macMenu, &itemNo);	
	
	if (MENUid != NOTHING) {			/* If command is in our table ...	*/
		i = FindMenuIndex(MENUid);
		HLock((Handle)theMenus);		/* Lock table and extract the 		*/
		theEntry = &(*theMenus)[i];		/*   entry of interest				*/
		
		DelMenuItem(macMenu, itemNo);	/* Remove item from Mac menu		*/
		
										/* Now remove command from our list	*/
		cmdList = *(theEntry->theCommands);
		
										/* Shift items up one slot			*/
		if (itemNo < theEntry->numCmds) {
			BlockMove(cmdList + itemNo, cmdList + itemNo -1,
					  (theEntry->numCmds - itemNo) * 4L);
		}

		theEntry->numCmds--;
		SetHandleSize((Handle) theEntry->theCommands, theEntry->numCmds * 4L);
		HUnlock((Handle) theMenus);
	}
}


/**** L O O K - U P   M E T H O D S ****/


/******************************************************************************
 FindMenuIndex

		Find the index number of a menu in the menu table
 ******************************************************************************/

short	CBartender::FindMenuIndex(
	short	MENUid)						/* MENU resource ID to search for	*/
{
	MenuEntryP			theEntry;		/* Pointer to menu table			*/
	register short		i;				/* Index into menu table			*/
	
	theEntry = *theMenus;				/* Dereference for speed			*/

	for (i = 0; i < numMenus; i++) {	/* Search thru all entries			*/
		if (theEntry[i].MENUid == MENUid) {
			return(i);					/* Match found. Exit here			*/
		}
	}
										/* We've fallen through the loop	*/
	return(-1);							/*   Menu is not in our table		*/
}	


/******************************************************************************
 FindMacMenu

		Find the Mac MenuHandle associated with a specified MENU ID. Return
		NULL if ID is not in our menu table.
 ******************************************************************************/

MenuHandle	CBartender::FindMacMenu(
	short	MENUid)						/* MENU resource ID to search for	*/
{
	short		index;					/* Index in menu table				*/
		
	index = FindMenuIndex(MENUid);		/* Search menu table for this ID	*/
	
										/* Non-negative index means ID is	*/
										/*   in our table					*/
	return ( (index >= 0) ? (*theMenus)[index].macMenu : NULL );
}


/******************************************************************************
 FindCmdNumber

		Find the command number corresponding to the specified Menu choice,
		which is indicated by a MENU resource ID and item number.
 ******************************************************************************/

long	CBartender::FindCmdNumber(
	short		MENUid,					/* Menu resource ID					*/
	short		itemNo)					/* Item number within the menu		*/
{
	MenuEntryP	theEntry;				/* Pointer to menu table			*/
	short		index;					/* Index into menu table			*/
	long		cmdNo;
	
	index = FindMenuIndex(MENUid);		/* Lookup menu in the menu table	*/
	
	if ( (index >= 0)  &&  (itemNo > 0) ) {
		theEntry = *theMenus;			/* Dereference for speed			*/

		if (itemNo <= theEntry[index].numCmds) {
										/* Use item number as index into	*/
										/*   list of commands for the menu	*/
			cmdNo = *(*theEntry[index].theCommands + itemNo - 1);
			if (cmdNo >= 0) {
				return(cmdNo);			/* Return positive command number	*/
			}
		}
										/* Menu has more items than command	*/
										/*   numbers or item has a negative	*/
										/*   command number. Put negative	*/
										/*   MENU id in the HiWord and the	*/
										/*   item number in the LoWord.		*/
		return(-(((long)MENUid) << 16) - itemNo);
		
	} else {							/* Menu is not in table or item		*/
		return(cmdNull);				/*   number is invalid				*/
										
	}
}
	

/******************************************************************************
 FindMenuItem

		Find the associated menu ID, menu handle, and item number for a
		specified command number
 ******************************************************************************/

void	CBartender::FindMenuItem(
	long		cmdNo,
	short		*MENUid,
	MenuHandle	*macMenu,
	short		*itemNo)
{
	register MenuEntryP	theEntry;		/* Menu table has one entry per		*/
										/*   menu							*/
	register short		i;				/* Index into menu table			*/
	register short		j;				/* Index into items in a menu		*/
	
	if (cmdNo < 0) {					/* Negative command numbers are a	*/
										/*   special case. HiWord is the	*/
		*MENUid = HiShort(-cmdNo);		/*   MENU id and LoWord is the item	*/
		*itemNo = LoShort(-cmdNo);		/*   number in the menu				*/
		*macMenu = (MenuHandle) GetResource('MENU', *MENUid);
	} else {
		theEntry = *theMenus;			/* Deference for speed				*/
	
										/* Each menu has an entry in table	*/
		for (i = 0; i < numMenus; i++) {
										/* Search thru all cmds for this	*/
										/*   menu							*/
			for (j = 0; j < theEntry[i].numCmds; j++) {
				if (cmdNo == *(*theEntry[i].theCommands + j)) {
										/* Match found!!					*/
					*MENUid = theEntry[i].MENUid;
					*macMenu = theEntry[i].macMenu;
					*itemNo = j + 1;
					return;				/* Take a quick exit				*/
				}
			}
		}
			
		*MENUid = NOTHING;				/* Falling thru the loops means		*/
		*macMenu = NULL;				/*   that no match was found		*/
		*itemNo = NOTHING;
	}
}
	

/******************************************************************************
 FindItemText

		Find the item number in a Menu which has the specified text. A
		value of zero (0) is returned if there is no such item or menu.
 ******************************************************************************/

short	CBartender::FindItemText(
	short		MENUid,
	Str255		itemStr)
{
	register MenuHandle		macMenu;
	short					numItems;
	register short			i;
	Str255					testStr;
	
	macMenu = FindMacMenu(MENUid);		/* Get Mac menu record given ID		*/
	
	if (macMenu != NULL) {				/* Make sure menu exists			*/
	
										/* Run thru all items in the menu	*/
		numItems = CountMItems(macMenu);
		for (i = 1; i <= numItems; i++) {
			GetItem(macMenu, i, testStr);
										/* Comparison is case-sensitive and	*/
										/*   diacritical marks matter		*/
			if (EqualString(itemStr, testStr, TRUE, TRUE)) {
				return(i);				/* Match found. Exit here			*/
			}
		}
	}
										/* Fell out of loop					*/
	return(0);							/*   No match found					*/
}


/**** C O M M A N D   E X T R A C T I O N ****/


/******************************************************************************
 ExtractCommands

		Command numbers are appended to the strings for names of
		menu items. We need to separate the item name and the command
		number and store the spearate entities in the proper places.
 ******************************************************************************/

void	CBartender::ExtractCommands(
	register MenuEntryP		theEntry)
{
	register short	itemNo;				/* Item number in the menu			*/
	Str255			itemStr;			/* Text of menu item				*/
	long			*cmdList;			/* List of menu commands			*/
	long			cmdNo;				/* A particular command number		*/
	
										/* Get number of items in menu		*/
	theEntry->numCmds = CountMItems(theEntry->macMenu);
	
	if (theEntry->numCmds == 0) {		/* No items means a custom menu or	*/
		theEntry->theCommands = NULL;	/*   menu items added at run time	*/
	
	} else {
										/* Create block to hold cmd numbers	*/
		theEntry->theCommands = (long**) NewHandleCanFail(theEntry->numCmds * 4L);
		FailNIL( theEntry->theCommands);
		
		HLock( (Handle) theEntry->theCommands);	/* Lock it down and dereference it	*/
		cmdList = *(theEntry->theCommands);
		
										/* Run thru all the items in menu	*/
		for (itemNo = 1; itemNo <= theEntry->numCmds; itemNo++) {
										/* Retrieve text of menu item		*/
			GetItem(theEntry->macMenu, itemNo, itemStr);
										/* Split up actual text of item and	*/
										/*   the associated command number	*/
			ParseItemString(itemStr, &cmdNo);
										/* Store command number in our list	*/
			cmdList[itemNo - 1] = cmdNo;
										/* Set menu item text sans appended	*/
										/*   command number					*/
			SetItem(theEntry->macMenu, itemNo, itemStr);
		}
		
		HUnlock( (Handle) theEntry->theCommands);
	}
}
	

/******************************************************************************
 ParseItemString

		Chop the command number off the end of the item string. Return
		both the shortened string and the command number.
 ******************************************************************************/

void	CBartender::ParseItemString(
	register Str255		itemStr,
	long				*cmdNo)
{
	register short		len;			/* Length of the item string		*/
	register Str255		cmdStr;			/* Holds characters of command no.	*/
	register short		i;				/* Index into item string			*/
	register short		j;				/* Index into command no. string	*/
	
	len = itemStr[0];					/* Length is in byte zero of string	*/
	cmdStr[0] = 0;						/* Make sure it's an empty string	*/
	
	for (i = 1; i <= len; i++) {		/* Search thru item string			*/
										/*   for a special delimiter		*/
		if (itemStr[i] == CMD_DELIMITER) {
										/* Found it. Now copy characters	*/
										/*   after delimiter into string	*/
										/*   for the command number			*/
			for (j = i+1; j <= len; j++) {
				cmdStr[j-i] = itemStr[j];
				itemStr[j] = 0;			/* Clear character from item string	*/
			}
			itemStr[i] = 0;				/* Clear out delimiter character	*/
			itemStr[0] = i - 1;			/* Reduce length of item string		*/
			cmdStr[0] = len - i;		/* Set length of command no. string	*/
			break;						/* Bust out of this for loop		*/
		}
	}
	
	StringToNum(cmdStr, cmdNo);			/* Convert cmd string to a number	*/
}
	

/******************************************************************************
 ExtractHierMenus

		Check a Toolbox menu record for hierarchical menus and automatically
		add them to the Toolbox menu list.
 ******************************************************************************/

void	CBartender::ExtractHierMenus(
	MenuHandle	macMenu,				/* Mac Menu record					*/
	short		index)					/* Index in menu table				*/
{
	register short	i;					/* Index into items in menu			*/
	short			itemCmd;			/* Command key equivalent for item	*/
	short			hMENUid;			/* MENU id for hierarchical menu	*/
	
		/* Run backwards thru items in the menu. Backwards because we	*/
		/* don't want to put the call to CountMItems as the exit		*/
		/* condition, which is evaluated each time thru the loop. The	*/
		/* order of insertion of hierarchical menus is not significant.	*/
		
	for (i = CountMItems(macMenu); i > 0; i--) {
	
										/* The command key equivalent flags	*/
										/*   hierarchical menus				*/
		GetItemCmd(macMenu, i, &itemCmd);
		if (itemCmd == hMenuCmd) {		/* This item has a hier menu		*/
										/*   attached. The MENU id for the	*/
										/*   hier menu is stored as the		*/
										/*   item mark.						*/
			(*theMenus)[index].hasHMenus = TRUE;
			GetItemMark(macMenu, i, &hMENUid);
			if (hMENUid != NOTHING) {	/* If a valid MENU id is specified,	*/
										/*   add it to the Bartender		*/
				AddMenu(hMENUid, TRUE, hierMenu);
			}
		}
	}
}


/**** I T E M   D I M M I N G / E N A B L I N G   M E T H O D S ****/


/******************************************************************************
 SetDimOption

		Specify the DimOption for a Menu. Items in all menus are dimmed
		according to this option whenever the menus are updated.

		Choices are:
			dimNONE - Don't dim any items in the menu
			dimSOME - Dim items with defined command numbers
						Items 1 thru numCmds
			dimALL  - Dim all items
 ******************************************************************************/

void	CBartender::SetDimOption(
	short		MENUid,
	DimOption	aDimming)
{
	short		index;
	
	index = FindMenuIndex(MENUid);		/* Find menu in our table			*/
	
	if (index >= 0) {					/* Set option if menu is in table	*/
		(*theMenus)[index].dimming = aDimming;
	}
}


/******************************************************************************
 SetUncheckOption

		Specify the UncheckOption for a Menu. Specifying TRUE means that
		all items in the menu are automatically unchecked whenever menus
		are updated.
 ******************************************************************************/

void	CBartender::SetUnchecking(
	short		MENUid,
	Boolean		anUnchecking)
{
	short		index;
	
	index = FindMenuIndex(MENUid);		/* Find menu in our table			*/
	
	if (index >= 0) {					/* Set option if menu is in table	*/
		(*theMenus)[index].unchecking = anUnchecking;
	}
}


/******************************************************************************
 UpdateAllMenus

		First dim all menu items, then let everyone in the active chain
		of command update their menu items
 ******************************************************************************/

void	CBartender::UpdateAllMenus()
{
	register MenuEntryP		theEntry;	/* Pointer to menu table			*/
	register short			i;			/* Index into menu table			*/
	register short			shift;
	unsigned long			mask;
	register short			j;
	short					numItems;
	Boolean					hasHMenus;
	MenuHandle				macMenu;
	short					itemCmd;
	
	theEntry = *theMenus;				/* Dereference for speed			*/

	for (i = 0; i < numMenus; i++) {	/* Run thru all menus				*/
		macMenu = theEntry[i].macMenu;
										/* Dim items according to the		*/
										/*   specific option for the menu	*/
		if (theEntry[i].dimming == dimALL) {
			(**macMenu).enableFlags &= 0x00000001;
			
		} else if (theEntry[i].dimming == dimSOME) {
										/* Dim items 1 thru numCmds			*/
			shift = theEntry[i].numCmds + 1;
			if (shift > 1) {			/* numCmds is 0 if shift is 1		*/
				if (shift > 31) {
					mask = 1L;			/* Only 31 items can be dimmed		*/
				} else {
					mask = (0xFFFFFFFF << shift) + 1;
				}
				(**macMenu).enableFlags &= mask;
			}
		}
		
		if (theEntry[i].unchecking) {
			numItems = CountMItems(macMenu);
			if (theEntry[i].hasHMenus) {
				for (j = 1; j <= numItems; j++) {
					GetItemCmd(macMenu, j, &itemCmd);
					if (itemCmd != hMenuCmd) {
						CheckItem(macMenu, j, FALSE);
					}
				}
			
			} else {
				for (j = 1; j <= numItems; j++) {
					CheckItem(macMenu, j, FALSE);
				}
			}
		}
	}

	gGopher->UpdateMenus();				/* Let active Bureaucrats update	*/
										/*   the menu items					*/
}	

/******************************************************************************
 UpdateMenuBar

	Determine if the menu bar needs to be redrawn. Calls DrawMenuBar if it does.
	The menu bar needs to be redrawn if the forceMBarUpdate flag is TRUE, or
	the enabled state of any menu in the menu bar doesn't match our last
	recorded state.
 ******************************************************************************/

void CBartender::UpdateMenuBar( void)
{
	MenuEntry	*mEntry;
	short		numMenus;
	Boolean		doIt = FALSE;
	
	if (forceMBarUpdate)
	{
		doIt = TRUE;
		forceMBarUpdate = FALSE;
	}
	else
	{
		mEntry = *theMenus;
		numMenus = this->numMenus;
		while (--numMenus >= 0)
		{
			if (mEntry->lastEnable != MenuEnabled( mEntry->macMenu))
			{
				mEntry->lastEnable = !mEntry->lastEnable;
				doIt = TRUE;
				
				// can't break here, because we still need to update
				// the lastEnable flag for each menu.
			}
			mEntry++;
		}
	}
	if (doIt) DrawMenuBar();
}