/******************************************************************************
 CCluster.c

							The Cluster Class
							
		An unordered group of objects.
	
	SUPERCLASS = CCollection
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
	TCL 1.1 CHANGES
	[
		- CCluster is now a superclass of CArray. See
		  CCluster.h for the interface changes.
		- implemented override of Copy
	]

 ******************************************************************************/

#include "Global.h"
#include "CCluster.h"


/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/
	

/******************************************************************************
 ICluster

		Initialize a Cluster object
 ******************************************************************************/

void	CCluster::ICluster()
{
	CArray::IArray( sizeof( CObject*));		/* Initialize superclass			*/
	
		/* set the items handle to be a copy of CArray::hItems	*/
		/* This helps to be compatible with existing subclasses */
		/* of CCluster and also lets us optimize certain things */
		/* because we know the item handle is really an array	*/
		/* of object references.								*/
		
	items = (LongHandle) hItems;
}
		
/******************************************************************************
 DisposeAll

		Dispose of all items contained in a Cluster and the Cluster itself.
 ******************************************************************************/

void	CCluster::DisposeAll()
{
	DisposeItems();
			
	this->Dispose();					/* Dispose of the Cluster itself	*/
}
	

/******************************************************************************
 DisposeItems

		Dispose of all items contained in a Cluster, but NOT the
		Cluster itself
 ******************************************************************************/

void	CCluster::DisposeItems()
{
	register long	i;					/* Index into Cluster				*/
	CObject			*itemToPurge;
	long	index;
	
	for (i = 0; i < numItems; i++) {	/* Run through items in the Cluster	*/
										/*   Send Dispose msg to each item	*/
		itemToPurge = (CObject*) (*items)[i];
		(*items)[i] = 0L;				/* Clear reference to disposed item	*/
		itemToPurge->Dispose();
		
		index = i+1;					/* notify dependents				*/
		BroadcastChange( arrayDeleteElement, &index);
	}
	
	numItems = 0;						/* There are now no items			*/
	if (slots >  blockSize) {			/* There are too many free slots	*/
		Resize( blockSize);				/* Reduce to one block of slots		*/
	}
	
}

	

/******************************************************************************
 Add

		Add a new object to the cluster
 ******************************************************************************/

void	CCluster::Add(
	CObject		*theObject)				/* Object to add					*/
{	
	InsertAtIndex( &theObject, numItems+1);	
}	


/******************************************************************************
 Remove

		Remove an object from the cluster
 ******************************************************************************/

void	CCluster::Remove(
	CObject		*theObject)				/* Object to remove					*/
{
	register long		i;				/* Offset of object in cluster		*/
	
	i = Offset(theObject);				/* Search items for this object		*/
	
	if (i != BAD_INDEX)
	{
		DeleteItem( i+1);
	}
}


/******************************************************************************
 Includes

		Check whether an object is in the cluster
 ******************************************************************************/

Boolean	CCluster::Includes(
	CObject		*theObject)				/* Object to check for inclusion	*/
{
										/* Object is in the cluster if the	*/
										/*   Offset is NOT BAD_INDEX		*/
	return(Offset(theObject) != BAD_INDEX);
}


/******************************************************************************
 DoForEach

		Call a specified function for each object in the cluster.
 ******************************************************************************/

void	CCluster::DoForEach(
	EachFunc	theFunc)				/* Function pointer					*/
{
	register long	i;
	Boolean wasLocked;
	
	TRY
	{
		wasLocked = SetLockChanges( TRUE);
		
		for (i = 0; i < numItems; i++) {
			(*theFunc)((CObject*) (*items)[i]);
		}
		
		SetLockChanges( wasLocked);
	}
	CATCH
	{
		SetLockChanges( wasLocked);
	}
	ENDTRY;
}


/******************************************************************************
 DoForEach1

		Call a specified function for each object in the cluster. The
		function to call has one parameter, which is a long.
 ******************************************************************************/

void	CCluster::DoForEach1(
	EachFunc1	theFunc,				/* Function pointer					*/
	long		param)					/* Parameter for the function		*/
{
	register long	i;
	Boolean wasLocked;
	
	TRY
	{
		wasLocked = SetLockChanges( TRUE);
	
		for (i = 0; i < numItems; i++) {
			(*theFunc)((CObject*)  (*items)[i], param);
		}
		SetLockChanges( wasLocked);
	}
	CATCH
	{
		SetLockChanges( wasLocked);
	}
	ENDTRY;
}


/******************************************************************************
 FindItem

		Find the first item in the cluster which successfully passes (returns
		TRUE) the test function. Return NULL if no item passes the test.
 ******************************************************************************/

CObject*	CCluster::FindItem(
	TestFunc		testFunc)			/* Ptr to test function				*/
{
	register long	i;					/* Index into list of items			*/
	Boolean 	wasLocked;
	CObject 	*item = NULL;
	
	TRY
	{
		wasLocked = SetLockChanges( TRUE);
	
		for (i = 0; i < numItems; i++) {	/* Scan from beginning to end		*/
			if ( (*testFunc)((CObject*)  (*items)[i]) )
			{
				item = (CObject*)(*items)[i];	/* Success! Pass back the object	*/
				break;
			}
		}
		SetLockChanges( wasLocked);	
	}
	CATCH
	{
		SetLockChanges( wasLocked);
	}
	ENDTRY;
	
	return item;
}


/******************************************************************************
 FindItem1

		Find the first item in the cluster which successfully passes (returns
		TRUE) the test function. Return NULL if no item passes the test.
 ******************************************************************************/

CObject*	CCluster::FindItem1(
	TestFunc1		testFunc,			/* Ptr to test function				*/
	long			param)				/* Parameter for test function		*/
{
	register long	i;					/* Index into list of items			*/
	Boolean 	wasLocked;
	CObject 	*item = NULL;
	
	TRY
	{
		wasLocked = SetLockChanges( TRUE);
	
		for (i = 0; i < numItems; i++) {	/* Scan from beginning to end		*/
			if ( (*testFunc)((CObject*)  (*items)[i], param) )
			{
				item = (CObject*)(*items)[i];	/* Success! Pass back the object	*/
				break;
			}
		}
		SetLockChanges( wasLocked);
	}
	CATCH
	{
		SetLockChanges( wasLocked);
	}
	ENDTRY;
	
	return item;	
}


/******************************************************************************
 Offset

		Find offset of an object within the cluster. Offset ranges
		from 0 to numItems-1.
 ******************************************************************************/

long	CCluster::Offset(
	CObject		*theObject)				/* Object to search for				*/
{
	register long	i;					/* Running index into the cluster	*/
	
	for (i = 0; i < numItems; i++) {	/* Search entire cluster			*/
										/* Return offset if object is found	*/
		if ( (long)theObject == (*items)[i] )
			return(i);
	}
										/* Falling thru the loop means that	*/
	return(BAD_INDEX);					/*   the search failed				*/
}

/******************************************************************************
 Copy {OVERRIDE}

	Duplicate a CCluster. This method calls CArray::Copy and the updates
	the items instance variable, an alias to CArray's hItems handle.
 ******************************************************************************/

CObject* CCluster::Copy( void)
{
	CCluster *theCopy;
	
	theCopy = (CCluster*) inherited::Copy();
	theCopy->items = (LongHandle) theCopy->hItems;
	
	return theCopy;
}