/******************************************************************************
 CDialog.c

	Implements the basic services needed for dialog windows. Dialogs windows
	have certain standard characteristics and behavior. They are either
	modal or modeless. There is usually a default button, which is hilighted
	visually and can be activated via the keyboard shortcut of Return or Enter.
	There is often a Cancel button, which can be activated via the keyboard
	shortcut of Command-'.' or the Escape key. When there is more than one item
	capable of responding to keyboard input, each may be activated by clicking
	on it with the mouse or the user may cycle through these items with the
	Tab key. As a new item becomes active, the previously active item becomes
	inactive.
		
	SUPERCLASS = CWindow
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CDialog.h"
#include "CDialogText.h"
#include "CButton.h"
#include "Constants.h"
#include "Commands.h"
#include "CScrollPane.h"
#include "CPaneBorder.h"
#include "CList.h"
#include "TBUtilities.h"

#define	kScrollSteps		10		/* default step value used when creating	*/
									/* a scrollpane for the dialog				*/

extern CBureaucrat *gGopher;

/******************************************************************************
 IDialog
 
 	Initialize a Dialog object from a WIND resource
 	
******************************************************************************/
void CDialog::IDialog(short WINDid, CDesktop *anEnclosure, CDirector *aSupervisor)
{
	CWindow::IWindow( WINDid, kNotFloating, anEnclosure, aSupervisor);
	IDialogX();

}	/* CDialog::IDialog */

/******************************************************************************
 INewDialog
 
 	Initialize a Dialog object dynamically from the given parameters
 	
******************************************************************************/
void CDialog::INewDialog( Rect *bounds, Boolean fVisible, short procID, Boolean fHasGoAway,
						CDesktop *anEnclosure, CDirector *aSupervisor)
{
						
	CWindow::INewWindow( bounds, fVisible, procID, kNotFloating, fHasGoAway,
				anEnclosure, aSupervisor);
	
	IDialogX();
	
}	/* CDialog::IDialog */

/******************************************************************************
 IDialogX
 
 	Common initialization for dialogs.
 	
******************************************************************************/
void CDialog::IDialogX(void)
{
	defaultBtn = NULL;
	scrollable = FALSE;
	itsPanorama = NULL;

}	/* CDialog::IDialogX */

/******************************************************************************
 MakePanorama 	
******************************************************************************/

void CDialog::MakePanorama( Boolean fHasHScroll, Boolean fHasVScroll,
						Boolean fHasSizeBox)

{
	CScrollPane *scrollPane = NULL;
	CView		*enclosure = this;
	LongRect		bounds;
	
	// only enable autscrolling if both scroll bars are present
	scrollable = fHasHScroll && fHasVScroll;

	if (fHasHScroll || fHasVScroll || fHasSizeBox)
	{
		scrollPane = new( CScrollPane);
		scrollPane->IScrollPane( this, this, 0, 0, 0, 0, sizELASTIC, sizELASTIC,
						fHasHScroll, fHasVScroll, fHasSizeBox);
		scrollPane->FitToEnclFrame( TRUE, TRUE);
		scrollPane->SetSteps( kScrollSteps, kScrollSteps);
		scrollPane->SetID( kDialogScrollPaneID);
		enclosure = scrollPane;
	} 
	itsPanorama = new( CPanorama);
	itsPanorama->IPanorama( enclosure, this, 0, 0, 0, 0, sizELASTIC, sizELASTIC);
	itsPanorama->FitToEnclosure( TRUE, TRUE);
	itsPanorama->SetWantsClicks( TRUE);
	
	if (scrollPane)
		scrollPane->InstallPanorama( itsPanorama);

	enclosure->GetInterior( &bounds);
	OffsetLongRect( &bounds, -bounds.left, -bounds.top);
	itsPanorama->SetBounds( &bounds);
	itsPanorama->SetID( kDialogPanoramaID);
	

}	/* CDialog::MakePanorama */

/******************************************************************************
 DoKeyDown {OVERRIDE}
 
 	Handle a keypress. Keypresses reach the dialog when they were not handled
 	by any other dialog item, i.e. the gopher. Here's where tabbing and the
 	basic keyboard shortcuts of return, enter, cmd-'.', and escape are handled.
 	
******************************************************************************/

void CDialog::DoKeyDown(char theChar, Byte keyCode, EventRecord *macEvent)
{
	switch (theChar)
	{
		case '\t':
		
			DoTab( (macEvent->modifiers & shiftKey) == 0);
			break;
			
		case '\r':
		case  kEnterKey:
			if (defaultBtn)
				defaultBtn->SimulateClick();
			break;

		default:
			if ((keyCode == KeyEscape) || IsCancelEvent( macEvent))
			{	
				CButton *cancelBtn = FindButton( cmdCancel);
				if (cancelBtn)
					cancelBtn->SimulateClick();
			}
			else
				inherited::DoKeyDown( theChar, keyCode, macEvent);
	}

}	/* CDialog::DoKeyDown */

/******************************************************************************
 DoTab
 
 	Handle a tab keypress by activating the next or previous dialog item. If 
 	current item is a CDialogText then it must be validated before deactivating
 	it.
 	
******************************************************************************/

void CDialog::DoTab( Boolean fForward)
{
	tGopherInfo	gopherInfo;
	CView *newGopher;
	
	gopherInfo.currentGopher = (CView*) gGopher;
	FindGophers( &gopherInfo);
	
	newGopher = fForward? gopherInfo.nextGopher : gopherInfo.prevGopher;
	if (newGopher && (newGopher != gGopher))
	{
		if (!newGopher->BecomeGopher( TRUE))
			return;
		
		if (member( newGopher, CAbstractText))
			((CAbstractText*)newGopher)->SelectAll( TRUE);
	}
		
}	/* CDialog::DoTab */

/******************************************************************************
 FindGophers
 
 	Search the view hierarchy of the dialog for potential gophers. The
 	current gopher, or NULL, should be entered in gophInfo.currentGopher
 	before calling FindGophers. The next, previous, first, and last
 	potential gophers are located in one fell swoop. The candidates
 	are selected on the basis of being visible and answering TRUE 
 	to CanBeGopher().
 	
******************************************************************************/

	static void View_FindGophers( CObject *aView, long gophInfo)
	{
		tGopherInfo *gopherInfo = (tGopherInfo*) gophInfo;
		CView *view = (CView*) aView;
		
		if (view->ReallyVisible())
		{
			if (view->CanBeGopher())
			{
				if (!gopherInfo->firstGopher)		/* firstGopher is NULL only before */
					gopherInfo->firstGopher = view;	/* we hit the first gopher */
					
				gopherInfo->lastGopher = view;		/* always update last gopher because */
													/* we can't tell which is last until */
													/* we're done 						*/
				
				if (gopherInfo->hitCurrent &&		/* after we hit the current gopher, */
						!gopherInfo->nextGopher)	/* the first one we find is the next */
					gopherInfo->nextGopher = view;	/* gopher.                          */
	
				if (view == gopherInfo->currentGopher) /* flag when we hit the current	*/
					gopherInfo->hitCurrent = TRUE;	/* gopher so we can watch for the	*/
													/* next one.						*/
				
				if (!gopherInfo->hitCurrent)		/* keep updating the prev gopher	*/
					gopherInfo->prevGopher = view; /* until we hit the current gopher */
			}	
			
			if (view->itsSubviews)				/* recursively search subviews 		*/
				view->itsSubviews->DoForEach1((EachFunc1) View_FindGophers, gophInfo);
				
		}
	}

void CDialog::FindGophers( tGopherInfo *gopherInfo)
{
	gopherInfo->firstGopher = NULL;
	gopherInfo->lastGopher = NULL;
	gopherInfo->nextGopher = NULL;
	gopherInfo->prevGopher = NULL;
	gopherInfo->hitCurrent = FALSE;
	
	if (itsSubviews)
	{
		itsSubviews->DoForEach1( (EachFunc1) View_FindGophers, (long) gopherInfo);
		
		if (!gopherInfo->nextGopher)	/* if no next gopher, wrap around to first */
			gopherInfo->nextGopher = gopherInfo->firstGopher;
			
		if (!gopherInfo->prevGopher)	/* if no prev gopher, wrap around to last */
			gopherInfo->prevGopher = gopherInfo->lastGopher;
	}

}	/* CDialog::FindGophers */

/******************************************************************************
 FindButton
 
 	Locates the first button in the view hierarchy associated with
 	the given command.
 	
******************************************************************************/

	static Boolean View_MatchCmd( CView *aView, long aCmd)
	{
		Boolean found;
		
		if (member( aView, CButton))
		{
			found = ((CButton*) aView)->GetClickCmd() == aCmd;
		}
		else
			found = FALSE;
			
		return found;
	}
	
CButton *CDialog::FindButton( long aCmd)
{
	return (CButton*) MatchView( (MatchViewProc) View_MatchCmd, (void*) aCmd);

}	/* CDialog::FindButton */

/******************************************************************************
 Close
 
	Close the dialog. The dialog is not closed if validation fails.
	 	
******************************************************************************/

void CDialog::Close(void)
{
	if (Validate())
	{
		inherited::Close();
	}

}	/* CDialog::Close */

/******************************************************************************
 Validate
 
 	Validates the dialog by validating all the CDialogText fields. Returns
 	FALSE if validation fails.
 	
******************************************************************************/

	static Boolean DialogText_Validate( CObject *aPane, void* ignored)
	{
		/* returns TRUE if aPane is a CDialogText and it failed validation */
		
		Boolean validationFailed = FALSE;

		if (member( aPane, CDialogText) && !((CDialogText*)aPane)->Validate())
		{
			validationFailed = TRUE;
		}

		return validationFailed;
	}

Boolean CDialog::Validate( void)
{
	CDialogText	*invalidText = NULL;
	
	if (itsSubviews)
	{
		invalidText = (CDialogText*) MatchView( (MatchViewProc) DialogText_Validate, NULL);
		
		if (invalidText)
		{
			invalidText->SelectAll( TRUE);
			invalidText->BecomeGopher( TRUE);
		}
	}
	
	return (invalidText == NULL);

}	/* CDialog::Validate */

/******************************************************************************
 SetDefaultCmd
 
 	Sets the default button to the one associated with the given command.
 	
******************************************************************************/

void CDialog::SetDefaultCmd( long aCmd)
{
	CButton *btn = FindButton( aCmd);
	
	if (btn)
	{
		SetDefaultButton( btn);
	}

}	/* CDialog::SetDefaultCmd */

/******************************************************************************
 SetDefaultButton
 
 	Sets the default button to aButton. The default button has the standard
 	bold outline and accepts the keyboard shortcut of return or enter.
 	
******************************************************************************/

void CDialog::SetDefaultButton( CButton *aButton)
{
	if (aButton != defaultBtn)
	{
		if (defaultBtn) defaultBtn->SetDefault( FALSE);
		if (aButton) aButton->SetDefault( TRUE);	
		defaultBtn = aButton;
	}
	
}	/* CDialog::SetDefaultButton */

/******************************************************************************
 SetCmdEnable
 
 	Enables or disables the button assocated with aCmd.
 	
******************************************************************************/

void CDialog::SetCmdEnable( long aCmd, Boolean fEnable)
{
	CButton *btn = FindButton( aCmd);
	
	if (btn)
	{
		if (fEnable) btn->Activate();
		else btn->Deactivate();
	}
	
}	/* CDialog::SetCmdEnable */

/******************************************************************************
 ScrollToPane
  	
  	Ensure that a pane is visible by scrolling it into view when necessary.
  	Does nothing if the dialog is not scrollable.
******************************************************************************/

void CDialog::ScrollToPane( CPane *aPane)
{
	LongRect	paneFrame;
	Rect		r;
	short		hSpan, vSpan;
	LongPt		selPos, position;
	CPaneBorder	*border;
	
	if (!scrollable) return;

		/* get the pane's frame and convert it to the coordinates		*/
		/* of the dialog panorama, which does the scrolling				*/
		
	aPane->GetFrame( &paneFrame);
	aPane->FrameToWindR( &paneFrame, &r);
	
	border = aPane->GetBorder();
	if (border)
		border->CalcBorderRect( &r);
	
	itsPanorama->WindToFrameR( &r, &paneFrame);
	itsPanorama->GetFrameSpan( &hSpan, &vSpan);
	
	itsPanorama->GetPosition( &position);
	selPos = position; 				/* get the current scroll position	*/
	
		/* grow the frame rectangle a little so we always have a little	*/
		/* whitespace around the pane									*/
		
	InsetLongRect( &paneFrame, -2, -2);
			
	/**** calc vertical scroll ****/
	
		/* Is the top of the pane below the bottom of the frame?		*/
	
	if (paneFrame.bottom >= position.v + vSpan)	
	{
		selPos.v = paneFrame.bottom - vSpan; 
	}
		
		/* Is the bottom of the pane above the top of the frame?		*/
	
	else if (paneFrame.top < position.v)
	{
		selPos.v = paneFrame.top;
	}
		
	/**** calc horizontal scroll ****/
	
		/* Is the left edge of the pane past the right edge of frame?	*/
		
	if (paneFrame.right >= position.h + hSpan)
	{
		selPos.h = paneFrame.right - hSpan;
	}
		
		/* Is the right edge of the pane before the left edge of frame? */
		 
	else if (paneFrame.left < position.h)
	{
		selPos.h = paneFrame.left;
	}
	
		/* scroll to new position of necessary							*/

	if ((selPos.h != position.h) || (selPos.v != position.v))
		itsPanorama->ScrollTo( &selPos, TRUE);

}	/* CDialog::ScrollToPane */

/******************************************************************************
  ProviderChanged
  
  	Respond to a change in a provider. All the dialog items are implicit
  	providers to the dialog. When an item becomes the gopher, it will notify
  	the dialog. If the gopher is a pane, the dialog will attempt to scroll
  	it into view.
******************************************************************************/

void CDialog::ProviderChanged( CCollaborator *aProvider, long reason, void* info)
{
	if (member( aProvider, CPane) && (reason == bureaucratIsGopher))
	{
		ScrollToPane( (CPane*) aProvider);
	}
	inherited::ProviderChanged( aProvider, reason, info);
	
}	/* CDialog::ProviderChanged */

