/******************************************************************************
 CPaneMDEF.c

							The PaneMDEF Class
							
	A menu definition procedure which treats the menu contents as a pane.
	Thus any pane which can be draw in a window can also be displayed in
	a menu. Also includes some support for tear-off menus.
			
	SUPERCLASS = CMenuDefProc
	
	Copyright  1989 Symantec Corporation. All rights reserved.

	Change History:
		01/30/90	pb	Added SetupQuickDraw and RestoreQuickDraw methods.

 ******************************************************************************/

#include "Global.h"
#include "CPaneMDEF.h"
#include "CTearOffMenu.h"
#include "CDesktop.h"
#include "CBartender.h"

#include <Script.h>

/*** Global Variables ***/

extern CDesktop		*gDesktop;			/* The visible Desktop				*/

/*** Class Constants ***/

#define		TEARMARGIN		10			/* Margin for tearing off a menu	*/
#define		TEARWAIT		6			/* Ticks to delay while dragging	*/
#define		FRAME_WIDTH		1			/* Thickness of window frame		*/
#define		DRAG_WIDTH		10			/* Thickness of window drag bar		*/


/******************************************************************************
 IPaneMDEF

		Initialize a PaneMDEF object
 ******************************************************************************/

void	CPaneMDEF::IPaneMDEF(
	short			MDEFid,
	CPane			*aPane,
	CTearOffMenu	*aTearOffMenu)
{
	CMenuDefProc::IMenuDefProc(MDEFid);	/* Initialize superclass			*/
	
	itsPane = aPane;
	itsTearOffMenu = aTearOffMenu;
}
	

/******************************************************************************
 DrawMenu {OVERRIDE}

		Draw the contents of the pane which will appear in the menu.
		Drawing is performed in the Desktop's port.
 ******************************************************************************/

void	CPaneMDEF::DrawMenu(
	MenuHandle		macMenu,
	Rect			*menuRect)
{
	LongRect		paneFrame;			/* Frame of pane to be drawn		*/
	Rect			qdFrame;

	SetupQuickDraw(menuRect);

	itsPane->GetFrame(&paneFrame);
	LongToQDRect( &paneFrame, &qdFrame);	
	itsPane->RestoreEnvironment();		/* Set up pane's drawing environs	*/	  
	itsPane->Draw(&qdFrame);			/* Draw pane inside the menu		*/
	
		/* Check if the entire menu is disabled.  If so, the menu	*/
		/* should be drawn in gray. We draw in gray by BIC'ing		*/
		/* a gray pattern over the entire pane.						*/
	
	if (!MenuEnabled(macMenu)) {		/* ??? Should selections be			*/
		PenMode(patBic);				/*     unhilited?					*/
		PenPat(gray);
		PaintRect( &qdFrame);
	}

	RestoreQuickDraw();
}
	

/******************************************************************************
 SizeMenu {OVERRIDE}

		Determine the height and width of the menu in pixels. Size of the
		menu is the same as that of the pane's frame.
 ******************************************************************************/

void	CPaneMDEF::SizeMenu(
	MenuHandle		macMenu)
{
										/* Get dimensions of its pane		*/
	itsPane->GetLengths(&(**macMenu).menuWidth, &(**macMenu).menuHeight);
}


/******************************************************************************
 TearOffMenu

		Drag dotted outline of torn off menu
 ******************************************************************************/

void	CPaneMDEF::TearOffMenu(
	register Rect	*menuRect,			/* Bounding box of menu				*/
	Point			mouseLoc)			/* Mouse location					*/
{
	Rect				margins;		/* Margins around menu rectangle	*/
	Rect				tearRect;		/* Current location of tear off		*/
	Pattern				grayPat;		/* Gray pattern for dragging		*/
	long				ticks;			/* Tick count after delay			*/
	
										/* Tear off occurs only when the	*/
										/*   mouse is in certain places		*/
	if (PtInTearRgn(mouseLoc, menuRect)) {
										/* Prepare to tear off the menu		*/
										/*   by dragging a gray outline		*/
										/*   of the menu					*/
		PenNormal();
		PenPat(gray);
		PenMode(patXor);				/* Toggle gray rect on the screen	*/
		SetClip(GetGrayRgn());
		
		itsTearOffMenu->GetMargins(&margins);
		do {							/* Seize control until the mouse	*/
										/*   is released or leaves the		*/
										/*   valid tear off region			*/
					
										/* Make a rectangle which is the	*/
										/*   same size as the window which	*/
										/*   holds the tear off menu		*/
			tearRect.left = menuRect->left - margins.left;
			tearRect.top = menuRect->top - margins.top;
			tearRect.right = menuRect->right + margins.right;
			tearRect.bottom = menuRect->bottom + margins.bottom;				
			
			OffsetRect(&tearRect, mouseLoc.h - menuRect->left,
								  mouseLoc.v - menuRect->top);
								  
				/* Drawing and then undrawing a gray outline at */
				/* the current location gives the illlusion of	*/
				/* dragging the tear off menu					*/

			FrameRect(&tearRect);		/* Draw the gray outline			*/
			Delay(TEARWAIT, &ticks);	/* Wait a while						*/
			FrameRect(&tearRect);		/* Undraw the gray outline			*/
				
			GetMouse(&mouseLoc);		/* Check current mouse location		*/
		} while (StillDown()  &&  PtInTearRgn(mouseLoc, menuRect));
		PenNormal();					/* Restore normal pen state			*/
		
										/* Check reason for loop exit		*/
		if (PtInTearRgn(mouseLoc, menuRect)) {
		
										/* Mouse released inside tear off	*/
										/*   region							*/
			itsTearOffMenu->TornOff(mouseLoc);
		}
	}
}


/******************************************************************************
 PtInTearRgn

		Check if a point is in a region where menu may be torn off
 ******************************************************************************/

Boolean	CPaneMDEF::PtInTearRgn(
	Point			hitPt,				/* Mouse location					*/
	register Rect	*menuRect)			/* Bounding box of menu				*/
{
		/* Valid tear off region:  Mouse is not in the menu bar and   */
		/* at least TEARMARGIN pixels from the left, right, or bottom */
		/* sides of the menu.										  */
		
	return( (hitPt.v > GetMBarHeight())  &&
			(
			  (hitPt.h < menuRect->left - TEARMARGIN)   ||
			  (hitPt.h > menuRect->right + TEARMARGIN)  ||
			  (hitPt.v > menuRect->bottom + TEARMARGIN)
			) 
		  );
}			

/******************************************************************************
 SetupQuickDraw

		Set up QuickDraw drawing environment so pane coordinates work.
 ******************************************************************************/

void	CPaneMDEF::SetupQuickDraw (Rect *menuRect)

{
	RgnHandle		tsaveClip;
	Point			newOrigin;			/* New origin for pane				*/
	Rect			desktopBounds;		/* Boundaries of the desktop		*/
	LongRect		paneFrame;
	Rect			qdFrame;

	GetPort(&savePort);					/* Save the current port			*/
	gDesktop->Prepare();				/* Draw into our Desktop's port		*/

	itsPane->GetFrame(&paneFrame);		/* Get size of pane and Desktop		*/
	LongToQDRect( &paneFrame, &qdFrame);
	gDesktop->GetBounds(&desktopBounds);
	
	tsaveClip = NewRgn();				/* Save clipping region of Desktop	*/
	GetClip(tsaveClip);
	saveClip = tsaveClip;

										/* Pane will be drawn in Desktop's	*/
										/*   port rather than its own, so	*/
										/*   we must adjust the pane's		*/
										/*   origin accordingly				*/
	newOrigin.h = paneFrame.left - menuRect->left + desktopBounds.left;
	newOrigin.v = paneFrame.top - menuRect->top + desktopBounds.top;
	SetOrigin(newOrigin.h, newOrigin.v);
	
	ClipRect(&qdFrame);				/* Pane's frame corresponds to the	*/
										/*   menu rectangle, so make sure	*/
										/*   we don't draw outside the menu	*/
	CView::cPreparedView = itsPane;

}


/******************************************************************************
 RestoreQuickDraw

		Restore QuickDraw drawing environment to saved values.
 ******************************************************************************/


void	CPaneMDEF::RestoreQuickDraw (void)

{
	Rect	desktopBounds;
	
	gDesktop->GetBounds(&desktopBounds);
										/* Restore origin and clipping		*/
										/*   region of our desktop			*/
	SetOrigin(desktopBounds.left, desktopBounds.top);
	SetClip(saveClip);
	DisposeRgn(saveClip);
	SetPort(savePort);
	CView::cPreparedView = NULL;

}
