/******************************************************************************
 CTearOffMenu.c

							The TearOffMenu Class
		
	Manages a menu and a Window which is displayed when the menu is torn off.
	
	SUPERCLASS = CDirector
	
	Copyright  1989 Symantec Corporation. All rights reserved.

 ******************************************************************************/

#include "Global.h"
#include "CTearOffMenu.h"
#include "CTearChore.h"
#include "CApplication.h"
#include "CFWDesktop.h"
#include "CWindow.h"

/*** Global Variables ***/

extern CApplication	*gApplication;		/* Application object				*/
extern CDesktop		*gDesktop;			/* The visible Desktop				*/
	

/******************************************************************************
 ITearOffMenu

		Initialize a TearOffMenu object
 ******************************************************************************/

void	CTearOffMenu::ITearOffMenu(    
    CApplication        *aSupervisor,   /* Boss in the chain of command     */
	short			WINDid)				/* Window template resource ID		*/
{
	CDirector::IDirector(aSupervisor);	/* Initialize superclass			*/
	
	itsWindow = new(CWindow);			/* Create a window to hold menu		*/
										/*   when it is torn off			*/
	itsWindow->IWindow(WINDid, TRUE, gDesktop, this);
	itsWindow->SetActClick(TRUE);

	itsPane = NULL;						/* Subclass must create the pane	*/
	corner.h = corner.v = 0;			/* Will be set when menu is torn	*/
	SetRect(&margins, 1, 1, 1, 1);		/* Space for a one pixel frame		*/
}
	

/******************************************************************************
 Suspend {OVERRIDE}

		Following the Human Interface Guidelines, tear-off menus are hidden
		when the application is suspended.
 ******************************************************************************/

void	CTearOffMenu::Suspend()
{
    if (active) { 
    
        itsWindow->HideSuspend();
        ((CFWDesktop *) gDesktop)->CalcTopFloat();
		active = TRUE;
	}
}
	

/******************************************************************************
 Resume {OVERRIDE}

		Resume by showing the tear-off window if it was visible before
		we were suspended.
 ******************************************************************************/

void	CTearOffMenu::Resume()
{
    if (active) {
    
        itsWindow->ShowResume();
        ((CFWDesktop *) gDesktop)->CalcTopFloat();
    }
}


/******************************************************************************
 CloseWind {OVERRIDE}

		Close the floating window by moving it offscreen so it is no
		longer visible
 ******************************************************************************/

void	CTearOffMenu::CloseWind(
	CWindow		*theWindow)
{
	theWindow->MoveOffScreen();
}
	

/******************************************************************************
 TornOff

		Menu has been torn off
 ******************************************************************************/

void	CTearOffMenu::TornOff(
	Point		aCorner)
{
	CTearChore		*theTearChore;

	corner = aCorner;					/* Store location of tear off		*/
	
	theTearChore = new(CTearChore);		/* Create and install an urgent		*/
	theTearChore->ITearChore(this);		/*   chore for moving the tear off	*/
										/*   menu to the new location		*/
	gApplication->AssignUrgentChore(theTearChore);
}
	

/******************************************************************************
 MoveToCorner

		Move the tear-off menu window to the location stored in the
		corner instance variable
 ******************************************************************************/

void	CTearOffMenu::MoveToCorner()
{
	itsWindow->Move(corner.h, corner.v);
	itsWindow->Select();
}
	

/******************************************************************************
 GetMacWindow

		Return a pointer to the Mac window record for the window associated
		with a TearOffMenu
 ******************************************************************************/

WindowPtr	CTearOffMenu::GetMacWindow()
{
	return(itsWindow->GetMacPort());
}
	

/******************************************************************************
 SetMargins

		Set the amount of margin between the window outline and the pane
		contained within the window. The margin includes the size of the
		frame and drag bar which form the structure region of the window
		used by a tear-off menu.
 ******************************************************************************/

void	CTearOffMenu::SetMargins(
	Rect		*aMargins)
{
	margins = *aMargins;				/* Set instance variable			*/
}
	

/******************************************************************************
 GetMargins

		Get the amount of margin between the window outline and the pane
		contained within the window.
 ******************************************************************************/

void	CTearOffMenu::GetMargins(
	Rect		*theMargins)
{
	*theMargins = margins;				/* Return instance variable			*/
}