/******************************************************************************
 CPNTGFile.c

							The PNTGFile Class
		
	Handles reading and writing PNTG files.
	
	SUPERCLASS = CDataFile
	
	Copyright  1989 Symantec Corporation. All rights reserved.

 ******************************************************************************/

#include "Global.h"
#include "CPNTGFile.h"
#include "TCLUtilities.h"


/*** Class Constants ***/

#define		PNTG_HEADER_SIZE	512		/* Bytes in header of a PNTG file	*/
#define		PNTG_PACK_SIZE		72		/* Number of bytes packed at once	*/
										/*   by PackBits. Each row is		*/
										/*   packed separately, so			*/
										/*     = PNTG_WIDTH / 8				*/


/******************************************************************************
 IPNTGFile

		Initialize a PNTGFile
 ******************************************************************************/

void	CPNTGFile::IPNTGFile()
{
	Boolean		savedAlloc;
	
    CDataFile::IDataFile();
										/* Allocate space for fixed-size	*/
										/*   header of a PNTG file			*/
										
	savedAlloc = SetAllocation( kAllocCanFail);
	header = NewHandleClear((long)PNTG_HEADER_SIZE);
	SetAllocation( savedAlloc);
	
	FailNIL( header);
}


/******************************************************************************
 Dispose {OVERRIDE}

		Dispose of a PNTGFile
 ******************************************************************************/

void	CPNTGFile::Dispose()
{
	if (header != NULL)					/* Throw out header					*/
		DisposHandle(header);

	inherited::Dispose();				/* Pass message on to superclass	*/
}


/******************************************************************************
 ReadNewBitMap

		Create a new BitMap object and read the contents of a PNTG file
		into it. Return a reference to the BitMap.
 ******************************************************************************/

CBitMap*	CPNTGFile::ReadNewBitMap(
	Boolean		makePort)				/* Make offscreen port for BitMap?	*/
{
	long		length;					/* Length of file in bytes			*/
	Ptr			packedBits = NULL;
	CBitMap		*theBitMap = NULL;
	short		scanLine;
	Ptr			srcPtr;
	Ptr			destPtr;
	Ptr			checkPtr;
	Boolean		savedAlloc;

	TRY
	{
		SetMark(0L, fsFromStart);			/* Read the fixed-size header		*/
		length = PNTG_HEADER_SIZE;
		HLock(header);
		ReadSome(*header, length);
		HUnlock(header);
				
		length = GetLength();					/* Find length of rest of file		*/
		length -= PNTG_HEADER_SIZE;	
		
		savedAlloc = SetAllocation( kAllocCanFail);
		packedBits = NewPtr(length);		/* Make buffer to store file		*/
											/*   contents						*/
		SetAllocation( savedAlloc);
		
		FailNIL( packedBits);
											/* File contains a compressed		*/
											/*   bit image						*/
		SetMark((long)PNTG_HEADER_SIZE, fsFromStart);
		
		ReadSome(packedBits, length);
	
		theBitMap = new(CBitMap);			/* Allocate a new bit map to hold	*/
											/*   uncompressed image. Images are	*/
											/*   always the same size.			*/
		savedAlloc = SetAllocation( kAllocCanFail);
		theBitMap->IBitMap(PNTG_WIDTH, PNTG_HEIGHT, makePort);
		SetAllocation( savedAlloc);

				
		srcPtr = packedBits;				/* Set up to uncompress the image	*/
		destPtr = theBitMap->macBitMap->baseAddr;
		
			/* PNTG files use PackBits to compress each scan line. We	*/
			/* must use UnpackBits to uncompress them. srcPtr points to	*/
			/* the current location in the compressed image. destPtr	*/
			/* points the current location in the bit map where the		*/
			/* uncompressed image is being stored. destPtr should		*/
			/* increment by PNTG_PACK_SIZE each time, since scan lines	*/
			/* have a fixed length. However, if the file is corrupted,	*/
			/* destPtr may point into outer space. We use checkPtr to	*/
			/* check the value of destPtr each time thru the loop.		*/
			
		checkPtr = destPtr;
		for (scanLine = 1; scanLine <= PNTG_HEIGHT; scanLine++) {
											/* UnpackBits automatically bumps	*/
											/*   up the src and dest pointers	*/
			UnpackBits(&srcPtr, &destPtr, PNTG_PACK_SIZE);
											/* Make sure everything went OK		*/
			if (destPtr != checkPtr + PNTG_PACK_SIZE) {
				break;						
			} else {
				checkPtr = destPtr;
			}
		}
		
		DisposPtr(packedBits);				/* No longer need compressed image	*/
	}
	CATCH
	{
		ForgetPtr( packedBits);
		ForgetObject( theBitMap);
	}
	ENDTRY;
	
	return(theBitMap);					/* Pass back the new BitMap			*/

}


/******************************************************************************
 WriteBitMap

		Write out contents of a BitMap into a PNTG file
 ******************************************************************************/

void	CPNTGFile::WriteBitMap(
	CBitMap		*theBitMap)
{
	long		length;
	Ptr			srcPtr;
	Ptr			destPtr;
	char		packedBits[PNTG_PACK_SIZE + 1];
	short		scanLine;
	
	SetMark(0L, fsFromStart);			/* Write out the fixed-size header	*/
	HLock(header);
	WriteSome(*header, (long)PNTG_HEADER_SIZE);
	HUnlock(header);


		/* PNTG file uses PackBits to compress each scan line in	*/
		/* bit image. We loop thru each row in the image, packing	*/
		/* the bits, then writing out the packed bits to the file.	*/

	srcPtr = theBitMap->macBitMap->baseAddr;
	for (scanLine = 1; scanLine <= PNTG_HEIGHT; scanLine++) {
		destPtr = packedBits;
		PackBits(&srcPtr, &destPtr, PNTG_PACK_SIZE);
		length = destPtr - packedBits;
		WriteSome(packedBits, length);
	}
	
	length = GetMark();
	SetLength(length);
									/* Force write to disk				*/
	FailOSErr( FlushVol("\p", volNum) );

}
