/******************************************************************************
 TCLUtilities.c

		Miscellaneous TCL utilities
		
		Copyright  1991 Symantec Corporation. All rights reserved.
				
 ******************************************************************************/

#include "TCLUtilities.h"
#include "Global.h"
#include "Constants.h"
#include "TBUtilities.h"
#include "CApplication.h"
#include "CError.h"
#include <Packages.h>

extern CApplication *gApplication;
extern CError		*gError;

/******************************************************************************
 ErrorAlert
 
 	Displays an alert for the given error and message codes. The error message
 	may be customized as follows:
 	
 	- If the low word of the message is non-zero, it is assumed to be
 	  the index of a string in a STR# resource.
 	  
 	   - if the high word of the message is zero, the STR# ID is the
 	     TCL's private STR# 301
 	     
 	   - if the high word is non-zero, it is taken as an offset from the
 	     base ID of 1024. For example, if the message is 655,363, or 0x000A0003 hex,
 	     the high word is 10 decimal and low word is 3 decimal. Since the high
 	     word is non-zero it is added to 1024. The TCL will do a 
 	     GetIndString( string, 1034, 3) to get the error string.
 	     
 	     The SpecifyMsg function in Exceptions.c will build the message longword
 	     for you. SpecifyMsg( 1034, 3) returns 655,363. See the comments for
 	     SpecifyMsg in Exceptions.c for more details.
 	
 	- if the low word of the message is zero the TCL looks for an 'Estr' 
 	  resource matching the error. If none is found, a generic error string 
 	  is used.
 	
 ******************************************************************************/
void ErrorAlert( short error, long message)
{
	Str255	errStr;
	Str63	numStr;
	short	strIndex, strID;
	short	alertID;
	
	errStr[0] = 0;					/* mark string empty */

	/*
	 *	First see if anyone filled in the
	 *	message field
	 */
	strIndex = LoShort( message);	/* string index goes in low word of message */
	if (strIndex > 0)
	{
		/* the STR# resource ID is either 131 or some value specified from	*/
		/* an implicit base value of 1024									*/
		strID = HiShort( message);
		if (strID == 0)
			strID = STR_TCLfailMsgs;
		else
			strID += kUserFailMsgBase;
		
		 
		GetIndString( errStr, strID, strIndex);
		
	}
	
	if (errStr[0] == 0)
	{
		/*
		 *	Next, see if there is a custom 'Estr' resource for the error.
		 *	If there isn't just get the default error string.
		 */
		 
		StringHandle	strH;
		 
		strH = (StringHandle) GetResource( ErrMsg_Res, error);
		if (!strH)
		 	strH = GetString( STRosError2);
		 if (strH)
		 	CopyPString( *strH, errStr);
	}
			
	NumToString( error, numStr);
	ParamText( errStr, numStr, NULL, NULL);
	
	if (!gApplication || (gApplication->GetPhase() != appRunning))
		alertID = ALRT_ExceptionAbort;
	else
		alertID = ALRT_Exception;
		
		// avoid infinite recursion in error handling by specifically
		// testing if the ALRT and DITL resources we need are there.
		
  	if ((GetResource('ALRT', alertID) == NULL) || (GetResource('DITL', alertID) == NULL))
  	{
  		if (gError)
   			gError->MissingResources();
   		else
   			ExitToShell();	// nothing else we can do...
   	}

	PositionDialog( 'ALRT', alertID);
	InitCursor();
	StopAlert( alertID, NULL);
}

/******************************************************************************
 NewHandleCanFail
 
 	Attempts to allocate a handle without drawing upon the memory reserve.
 
******************************************************************************/
void *NewHandleCanFail( long size)
{
	Handle h;
	Boolean savedAlloc;
	
	ASSERT( size >= 0);
	
	savedAlloc = SetAllocation( kAllocCanFail);
	h = NewHandle( size);
	SetAllocation( savedAlloc);
	
	return h;
}

/******************************************************************************
 ResizeHandleCanFail
 
 	Attempts to resize a handle without drawing upon the memory reserve.
 	
******************************************************************************/
void ResizeHandleCanFail( void* theHandle, long newSize)
{	
	Boolean savedAlloc;
	
	ASSERT( newSize >= 0);
	
	savedAlloc = SetAllocation( kAllocCanFail);
	SetHandleSize( theHandle, newSize);
	SetAllocation( savedAlloc);

	/* MemError() now has error code, caller can FailMemError()	*/	
}

/******************************************************************************
 SetAllocation
 	
 	Changes the parameters gApplication uses when the grow zone function is
 	invoked. This happens when the Memory Mgr can't satisfy a requested
 	memory allocation. This method also returns the previous setting, 
 	so you can do something like:
 	
 		Boolean oldAlloc = SetAllocation( kAllocCanFail);
 		
 			...attempt something that could fail gracefully...
 		
 		SetAllocation( oldAlloc);
 		
 	If canFail is TRUE, this function does gApplication->RequestMemory( no loan, can fail)
 	otherwise it does gApplication->RequestMemory( no loan, can't fail),
 	i.e. this utility always sets loan approved FALSE.
 		
******************************************************************************/
Boolean	SetAllocation( Boolean canFail)
{
	Boolean oldFail = gApplication->canFail;
	
	gApplication->RequestMemory( canFail);
	
	return oldFail;
}

/******************************************************************************
 SetCriticalOperation
 
	Changes the parameters gApplication uses when the grow zone function
	is called. If the application is in a critical operation, then more 
	of the memory reserve is eligible to be released to satisfy a memory
	request.
 	
******************************************************************************/
void SetCriticalOperation( Boolean aCriticalOp)
{
	gApplication->SetCriticalOperation( aCriticalOp);
}

/******************************************************************************
 cstrlen
 
 	Return the length of a cstring. It's here so we don't have to drag in ANSI.
******************************************************************************/

long cstrlen( register char *s )
{
	register long count = 0;
	
	while (*s++) count++;
	return count;
}

/******************************************************************************
 __ForgetHandle
 
 	If the handle isn't already NULL, it disposes of the handle and NULLs it.
******************************************************************************/

pascal void __ForgetHandle( Handle *handleAddr)
{
	Handle h = *handleAddr;
	if (h)
	{
		*handleAddr = NULL;
		DisposHandle( h);
	}
}

/******************************************************************************
 __ForgetObject
 
 	If the object isn't already NULL, it sends it a Dispose message, then
 	NULLs it.
******************************************************************************/

pascal void __ForgetObject( Ptr *objAddr)
{
	CObject *obj = (CObject*) *objAddr;
	
	if (obj)
	{
		*objAddr = NULL;
		obj->Dispose();
	}
}

/******************************************************************************
 __ForgetPtr
 
 	If the pointer isn't already NULL, it disposes of the pointer and NULLs it.
******************************************************************************/

pascal void __ForgetPtr( Ptr *ptrAddr)
{ 
	Ptr p = *ptrAddr;
	
	if (p)
	{
		*ptrAddr = NULL;
		DisposPtr( p);
	}
}

/******************************************************************************
 __ForgetResource
 
  	If the resource isn't already NULL, it releases and NULLs it.
******************************************************************************/

pascal void __ForgetResource( Handle *rsrcAddr)
{
	Handle h = *rsrcAddr;
	
	if (h)
	{
		*rsrcAddr = NULL;
		if (HomeResFile(h) > 1)
			ReleaseResource( h);
	}
}

/******************************************************************************
 EqualMem
 
 	Compares two blocks of memory and returns TRUE if they are equal. Its
 	interface is:
 	pascal Boolean	EqualMem( void *p1, void *p2, long n);
 	
 	where p1 and p2 are pointers to two blocks of memory and n is the
 	number of bytes to compare.

******************************************************************************/

static pascal void AsmEqualMem(void)
{
	asm
	{
extern EqualMem:

			movem.l	(a7)+, d0/d1/a0/a1		; d0 = return address, d1 = n
			move.b	#1, (a7)				; assume equal
			subq.l	#1, d1
			beq.s	@byteCompare
			lsr.l	#1, d1

wordCompare:
			cmpm.w	(a0)+, (a1)+
			dbne		d1, @wordCompare
			beq.s	@wordExit				
			move.b	#0, (a7)   
wordExit:
			movea.l	d0, a0
			jmp		(a0)
		
byteCompare:
			cmpm.b	(a0)+, (a1)+
			beq.s	@byteExit
			move.b	#0, (a7)
byteExit:
			movea.l	d0, a0
			jmp		(a0)
	}
}

/******************************************************************************
 SetMinimumStack

  	Sets the stack to at least minSize. Only call this once, as the very
	first statement in your program
******************************************************************************/

void SetMinimumStack( long minSize)
{
	long newApplLimit;
	
	if (minSize > DefltStack)
	{
		newApplLimit = (long) GetApplLimit() - (minSize - DefltStack);
		SetApplLimit( (Ptr) newApplLimit);
	}
}

#ifdef __TCL_DEBUG__
/******************************************************************************
 __noObject
 
 	Called by oopsDebug if message is sent to a bad object pointer/handler
 ******************************************************************************/

void __noObject(void)
{
	Failure( paramErr, excMsgNullObject);
}
/******************************************************************************
 __noMethod
 
 	Called by oopsDebug if a message lookup fails
 ******************************************************************************/

void __noMethod(void)
{
	Failure( paramErr, excMsgLookupFailed);
}
#endif
