/******************************************************************************
 CStyleTEEditTask.c

		A subclass of CTextEditTask that handles undo for typing, cut, copy,
		paste, and clear commands. It works directly with TextEdit's
		styled scrap data structures to save and restore formatting.
		
		If you are using the CStyleText class then a CStyleTEEditTask
		is automatically created for you when you the user types or
		executes a clipboard command.
		
	SUPERCLASS = CTextEditTask
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CStyleTEEditTask.h"
#include "TCLUtilities.h"
#include "CClipboard.h"
#include "Global.h"


extern CClipboard *gClipboard;

/******************************************************************************
 IStyleTEEditTask
 
 	Initialize the CStyleTEEditTask. anEditCmd should be the command that the
	task is to respond to, i.e. cmdCut, cmdCopy, cmdPaste, or cmdClear. If
	the task is for typing, then it should be cmdNull. FirstTaskIndex is 
	the index into the strUNDO 'STR#' where the text undo labels are located.
	These are typically "Typing", "Cut", "Copy", "Paste", "Clear", and 
	"Formatting". This method also saves the current 'styl' data on the
	clipboard, in any.
	
******************************************************************************/

void CStyleTEEditTask::IStyleTEEditTask( CStyleText *aTextPane, long anEditCmd, 
					short firstTaskIndex)
{
	Handle	h;
	
	origStyleScrap = deletedStyles = insertedStyles = NULL;
	macTE = ((CStyleText*) aTextPane)->macTE;

	CTextEditTask::ITextEditTask( aTextPane, anEditCmd, firstTaskIndex);
	
	gClipboard->GetData( 'styl', &h);
	origStyleScrap = (StScrpHandle) h;

}	/* CStyleTEEditTask::IStyleTEEditTask */

/******************************************************************************
 SaveRange {OVERRIDE}
 
 	Called either to save the selected text before it is deleted, or the
 	inserted text before an undo is performed. This method calls the
 	inherited::SaveRange to do most of the work, and the saves the style
 	scrap data as well.
 	
******************************************************************************/

void CStyleTEEditTask::SaveRange( tRangeSelector whichRange)
{
	StScrpHandle	h;
	long oldStart, oldEnd;
	
	inherited::SaveRange( whichRange);
	
	itsTextPane->GetSelection( &oldStart, &oldEnd);
	
	if (whichRange == kInsertedRange)
		itsTextPane->SetSelection( inserted.start, inserted.end, FALSE);
	else
		itsTextPane->SetSelection( deleted.start, deleted.end, FALSE);
		
	h = ((CStyleText*)itsTextPane)->GetStyleScrap();
	
	itsTextPane->SetSelection( oldStart, oldEnd, FALSE);
	
	if (whichRange == kInsertedRange)
		insertedStyles = h;
	else
		deletedStyles = h;
	
}	/* CStyleTEEditTask::SaveRange */

/******************************************************************************
 RestoreRange {OVERRIDE}
 
 	Restores the previously saved range of either inserted or deleted text.
 	This method does not call inherited::RestoreRange.
******************************************************************************/

void CStyleTEEditTask::RestoreRange( tRangeSelector whichRange, Boolean killData)
{
	Handle			text;
	StScrpHandle 	styles;
	long			selStart, selEnd, start;
	short			styleFixup;
	
	if (whichRange == kInsertedRange)
	{
		text = inserted.text;
		styles = insertedStyles;
		start = inserted.start;
		selStart = inserted.selStart;
		selEnd = inserted.selEnd;
		if (killData)
		{
			inserted.text = NULL;
			insertedStyles = NULL;
		}
	}
	else
	{
		text = deleted.text;
		styles = deletedStyles;
		start = deleted.start;
		selStart = deleted.selStart;
		selEnd = deleted.selEnd;
		if (killData)
		{
			deleted.text = NULL;
			deletedStyles = NULL;
		}
	}
	
	if (text)
	{
		// Due to backspacing, the save styled scrap handle may
		// begin with negative offsets. Before attempting to
		// restore these styles we must fix up the offsets
		// to start at 0.
		
		styleFixup = -(**styles).scrpStyleTab[0].scrpStartChar;
		if (styleFixup > 0)
		{
			ScrpSTElement	*table = (**styles).scrpStyleTab;
			short i = (**styles).scrpNStyles;
			
			while (--i >= 0)
			{ 
				table++->scrpStartChar += styleFixup;
			}
		}

		itsTextPane->SetSelection( start, start, TRUE);
		MoveHHi( text);
		HLock( text);
		((CStyleText*)itsTextPane)->InsertWithStyles( *text, 
								GetHandleSize( text), styles);
		HUnlock( text);
	}
	if (killData)
	{
		ForgetHandle(text);
		ForgetHandle( styles);
	}
	itsTextPane->SetSelection( selStart, selEnd, TRUE);

}	/* CStyleTEEditTask::RestoreRange */

/******************************************************************************
 StoreToClip {OVERRIDE}
******************************************************************************/

void CStyleTEEditTask::StoreToClip( tClipSelector whichClip)
{
	StScrpHandle 	h;
	OSErr			err;

	inherited::StoreToClip( whichClip);
		
	if (whichClip == kOldClip)
		h = origStyleScrap;
	else
		h = deletedStyles;

	if (h)
		gClipboard->PutData( 'styl', (Handle) h);
		

}	/* CStyleTEEditTask::StoreToClip */

/******************************************************************************
 DoBackspace {OVERRIDE}
 
 	Handle a backspace over original (not newly typed) text. This method
 	checks whether the style of the char about to be deleted is different
 	from the style at the start of the saved styled scrap and saves it if
 	necessary.
******************************************************************************/

void CStyleTEEditTask::DoBackspace( void)
{
	long 			selStart, selEnd;
	TextStyle 		style;
	ScrpSTElement 	scrapEl;
	short			lineHeight, fontAscent;
	
	itsTextPane->GetSelection( &selStart, &selEnd);
	
	/*
	 *	We only need to do something when the user is backspacing
	 *  over original text that was not selected at the start
	 *  of this task. When that happens, we must check the style
	 *  of the char being deleted. If it is different from the
	 *  style currently at the front of the deletedStyle handle
	 *  then we must insert it to the handle.
	 */
	
	if ((selStart > 0) && (selStart == deleted.start) && (selStart == selEnd))
	{
		TEGetStyle( selStart-1, &style, &lineHeight, &fontAscent, macTE);
		
		/* make up a style scrap element for the char to be deleted.	*/
		/* Then we'll see if it matches the first style in the			*/
		/* saved scrap. If not, we need to insert it to the scrap		*/
		
		scrapEl.scrpStartChar = (**deletedStyles).scrpStyleTab[0].scrpStartChar-1;
		scrapEl.scrpHeight = lineHeight;
		scrapEl.scrpAscent = fontAscent;
		/* remaining fields are equivalent to a TextStyle record	*/
		*(TextStyle*) &scrapEl.scrpFont = style;
		
		/* CheckNewStyle inserts the new style at the head of the	*/
		/* style scrap if its new. Otherwise, we just decrement		*/
		/* the starting offset for the first style. This is			*/
		/* faster than running through the whole table, fixing		*/
		/* up offsets each time the user backspaces, but we have	*/
		/* to be sure and fix these up later, before attempting to	*/
		/* use these styles											*/
		
		if (!CheckNewStyle( &scrapEl, deletedStyles, TRUE))		
			(**deletedStyles).scrpStyleTab[0].scrpStartChar--;
	}

	inherited::DoBackspace();
}

/******************************************************************************
 DoFwdDelete {OVERRIDE}
 
 	Handle a forward delete over original text. This method
 	checks whether the style of the char about to be deleted is different
 	from the style at the end of the saved styled scrap and saves it if
 	necessary.
 	
******************************************************************************/

void CStyleTEEditTask::DoFwdDelete( void)
{
	long selStart, selEnd, length;
	TextStyle 		style;
	ScrpSTElement 	scrapEl;
	short			lineHeight, fontAscent;
	
	itsTextPane->GetSelection( &selStart, &selEnd);
	length = itsTextPane->GetLength();
	
	if ((selStart == selEnd) && (selEnd < length))
	{
		TEGetStyle( selStart, &style, &lineHeight, &fontAscent, macTE);

		/* make up a style scrap element for the char to be deleted.	*/
		/* Then we'll see if it matches the last style in the			*/
		/* saved scrap. If not, we need to insert it to the scrap		*/
		
		scrapEl.scrpStartChar = deleted.end - deleted.start + 
					(**deletedStyles).scrpStyleTab[0].scrpStartChar;
		scrapEl.scrpHeight = lineHeight;
		scrapEl.scrpAscent = fontAscent;
		/* remaining fields are equivalent to a TextStyle record	*/
		*(TextStyle*) &scrapEl.scrpFont = style;
		
		CheckNewStyle( &scrapEl, deletedStyles, FALSE);	

	}
	inherited::DoFwdDelete();
}

/******************************************************************************
 Dispose {OVERRIDE}
 
 	Dispose of this task.
******************************************************************************/

void CStyleTEEditTask::Dispose( void)
{
	ForgetHandle( origStyleScrap);
	ForgetHandle( deletedStyles);
	ForgetHandle( insertedStyles);
	inherited::Dispose();
}

/******************************************************************************
 CheckNewStyle
 
 	Utility method to compare a ScrpSTEElement with either the first or
 	last element in a StScrpHandle. If they are the same, FALSE is returned.
 	If they are different, the element is inserted in the handle and TRUE is
 	returned.
******************************************************************************/

Boolean CStyleTEEditTask::CheckNewStyle( ScrpSTElement *scrapEl, 
							StScrpHandle styleH, Boolean atStart)
{
	short 	index = atStart? 0 : (**styleH).scrpNStyles-1;
	long	currSize;
	Boolean wasNew = FALSE;
	
	/* compare scrapEl with either first or last scrapElement in scrap handle	*/
	/* but don't compare the scrpStartChar fields								*/
	/* If they don't match, then a new ScrpSTElement is inserted				*/
	
	if (!EqualMem( &scrapEl->scrpHeight, 
			&(**styleH).scrpStyleTab[index].scrpHeight, 
			sizeof(TextStyle) - sizeof(long)))
	{
			
		currSize = GetHandleSize( (Handle) styleH);
		ResizeHandleCanFail( (Handle) styleH, currSize + sizeof(ScrpSTElement));
		FailMemError();
		
		/* if atStart, then move old styles down			*/
		
		if (atStart)
		{
			BlockMove( &(**styleH).scrpStyleTab[0],
					&(**styleH).scrpStyleTab[1],
					currSize - sizeof(short));
		}
		else
			index++; // place at end
		
		/* place new style in position, increment count		*/
		
		(**styleH).scrpStyleTab[index] = *scrapEl;
		(**styleH).scrpNStyles++;
				
		wasNew = TRUE;
	}

	return wasNew;
}