/******************************************************************************
 CStyleText.c

	A subclass of CEditText that uses Styled TextEdit to allow multiple
	fonts, size, and styles within the text. The text pane still may have
	only a single alignment. CStyleText only supports single-spaced text.
	
	When using this class, you must also use CStyleTEEditTask and CStyleTEStyleTask.
	These classes provide undo support for typing, clipboard, and formatting
	commands. These tasks are automatically created for you via overrides of
	MakeEditTask and MakeStyleTask. 
	
	There is also a subclass of CClipboard,  CStyleTEClipboard, that can 
	display styled text on the scrap. You should override MakeClipboard 
	in your CApplication subclass if you want to use CStyleTEClipboard.
	
	Note that CStyleText uses pixels for scroll units, not lines.
	
	SUPERCLASS: CEditText
	
		Copyright  1991 Symantec Corporation. All rights reserved.
		
 ******************************************************************************/

#include "CStyleText.h"
#include "CPrinter.h"
#include "CScrollPane.h"
#include "Script.h"
#include "CStyleTEEditTask.h"
#include "CStyleTEStyleTask.h"
#include "CClipboard.h"
#include "Commands.h"
#include "TCLUtilities.h"
#include "Global.h"


#define kHScrollStep	10		/* horizontal scroll unit	*/

extern CClipboard *gClipboard;


/******************************************************************************
 IStyleText
 
 	Initialize a CStyleText pane.
******************************************************************************/

void CStyleText::IStyleText(CView *anEnclosure, CBureaucrat *aSupervisor,
						short aWidth, short aHeight,
						short aHEncl, short aVEncl,
						SizingOption aHSizing, SizingOption aVSizing,
						short aLineWidth)
{
	CEditText::IEditText( anEnclosure, aSupervisor, aWidth, aHeight,
				aHEncl, aVEncl, aHSizing, aVSizing, aLineWidth);
		
	IStyleTextX();			
}

/******************************************************************************
 IViewTemp
 
 	Initialize a CStyleText pane from a resource template.
******************************************************************************/

void CStyleText::IViewTemp(CView *anEnclosure, CBureaucrat *aSupervisor,
							Ptr viewData)
{
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);
	IStyleTextX();
}

/******************************************************************************
 IStyleTextX
 
 	Perform common initialization for CStyleText panes.
******************************************************************************/

void CStyleText::IStyleTextX( void)
{	
	fixedLineHeights = FALSE;
	SetScales( 1, 1);
	SetWholeLines( FALSE);
	AdjustBounds();

}

/******************************************************************************
 MakeMacTE {OVERRIDE}
 
 	This method creates a styled TextEdit record.
******************************************************************************/

void CStyleText::MakeMacTE( void)
{
	LongRect	longInterior;
    Rect        tFrame;
    Boolean		savedAlloc;

	GetInterior( &longInterior);
	LongToQDRect( &longInterior, &tFrame);
	
	savedAlloc = SetAllocation( kAllocCanFail);
    macTE = TEStylNew(&tFrame, &tFrame);
    SetAllocation( savedAlloc);
    
    FailNIL( macTE);
}

/******************************************************************************
 SetScrollPane {OVERRIDE}
 
 	Ensures that reasonable scroll steps are setup when a CStyleText
 	pane is installed in a scroll pane.
******************************************************************************/

void CStyleText::SetScrollPane( CScrollPane *aScrollPane)
{
	inherited::SetScrollPane( aScrollPane);
	
	aScrollPane->SetSteps( kHScrollStep, GetDefFontSize());
}

/******************************************************************************
 PerformEditCommand {OVERRIDE}
 
 	Handles the edit commands at the request of a CStyleTEEditTask. This method
 	does the right thing so both 'TEXT' and 'styl' data can be cut and
 	pasted from/to the scrap.
******************************************************************************/

void CStyleText::PerformEditCommand( long theCommand)
{
	Prepare();

	switch( theCommand)
	{		
		case cmdCut:
			gClipboard->EmptyGlobalScrap();			
			TECut(macTE);
			gClipboard->UpdateDisplay();
			break;
			
		case cmdCopy:
			gClipboard->EmptyGlobalScrap();
			TECopy(macTE);
			gClipboard->UpdateDisplay();
			break;
			
		case cmdPaste:
			CheckInsertion( gClipboard->DataSize( 'TEXT'), TRUE);
			TEStylPaste(macTE);
			break;
			
		case cmdClear:
			TEDelete(macTE);
			break;
			
	}
	AdjustBounds();
	ScrollToSelection();
}

/******************************************************************************
 InsertWithStyles
 
 	Insert text with associated style information into a CStyleText pane.
******************************************************************************/

void CStyleText::InsertWithStyles( Ptr text, long length, StScrpHandle styles)
{
	CheckInsertion( length, FALSE);
	
	Prepare();
	TEStylInsert( text, length, styles, macTE);
	AdjustBounds();
}

/******************************************************************************
 SetStyle {OVERRIDE}
 
 	Sets the style of the selection (or insertion point) to the given
 	style. mode and newStyle work as described in Inside Mac V and VI.
******************************************************************************/

void CStyleText::SetStyle( short mode, TextStyle *newStyle, Boolean redraw)
{
	Prepare();
	TESetStyle( mode, newStyle, redraw, macTE);
	TECalText( macTE);
	SetSpacingCmd( spacingCmd);
	SetWholeLines( wholeLines);
	AdjustBounds();
}

/******************************************************************************
 SetFontNumber {OVERRIDE}
 
 	Set the font of the selection.
******************************************************************************/

void CStyleText::SetFontNumber(short aFontNumber)
{
	TextStyle	style;
	
	style.tsFont = aFontNumber;
	SetStyle( doFont, &style, TRUE);
}

/******************************************************************************
 SetFontStyle {OVERRIDE}
 
 	Set the style of the selection
******************************************************************************/

void CStyleText::SetFontStyle(short aStyle)
{
	TextStyle	style;
	short		mode = doFace;
	
	style.tsFace = aStyle;
	if (aStyle != NOTHING)
		mode += doToggle;
	SetStyle( mode, &style, TRUE);

}

/******************************************************************************
 SetFontSize {OVERRIDE}
 
 	Set the font size of the selection
******************************************************************************/

void CStyleText::SetFontSize( short aSize)
{
	TextStyle	style;
	
	style.tsSize = aSize;
	SetStyle( doSize, &style, TRUE);

}

/******************************************************************************
 AdjustBounds {OVERRIDE}
 
 	Keeps the bounds instance variable and scroll pane in synch with the
 	extent of the text, which may change when the window is resized, or the
 	text is edited or reformatted.
******************************************************************************/

void CStyleText::AdjustBounds( void)
{
	register TEPtr	t;
	short  	newHeight;
	
	newHeight = GetHeight( 0, MAXINT);
	t = *macTE;
	
	bounds.left = bounds.top = 0;
	bounds.bottom = newHeight;

	if (lineWidth > 0) 
	{
		bounds.right = lineWidth;
	} else {
		bounds.right = t->viewRect.right - t->viewRect.left;
	}
        
    bounds.right = (bounds.right - 1) / hScale + 1;

	if (itsScrollPane != NULL) {
		itsScrollPane->AdjustScrollMax();
	}
}

/******************************************************************************
 SetSpacingCmd {OVERRIDE}
 
 	CStyleText does not support variable spacing.
 ******************************************************************************/

void CStyleText::SetSpacingCmd( long aSpacingCmd)
{
	spacingCmd = cmdSingleSpace;
	
	// only single-spaced text is supported.
	
  SetWholeLines(wholeLines);
  CalcAperture();
  AdjustBounds();

}

/******************************************************************************
 MakeEditTask {OVERRIDE}
 
 	Creates a CStyleTEEditTask to handle typing or edit menu commands.
******************************************************************************/

CTextEditTask *CStyleText::MakeEditTask( long editCmd)
{
	CStyleTEEditTask *editTask = NULL;
	
	TRY
	{
		editTask = new( CStyleTEEditTask);
		editTask->IStyleTEEditTask( this, editCmd, cFirstTaskIndex);
		
	}
	CATCH
	{
		ForgetObject( editTask);
	}
	ENDTRY;
	
	return editTask;
}

/******************************************************************************
 MakeStyleTask {OVERRIDE}

	Create a task to carry out font, style, size, alignment, and spacing
	commands. Must be a subclass of CTextStyleTask.
 ******************************************************************************/

CTextStyleTask *CStyleText::MakeStyleTask( long styleCmd)
{
	CStyleTEStyleTask *newTask = NULL;
	short			taskIndex;
	
	TRY
	{
		taskIndex = cFirstTaskIndex > 0 ? cFirstTaskIndex + undoFormatting : 0;
		newTask = new( CStyleTEStyleTask);
		
		newTask->IStyleTEStyleTask( this, styleCmd, taskIndex);
	}
	CATCH
	{
		ForgetObject( newTask);
	}
	ENDTRY;
	
	return newTask;
}

/******************************************************************************
 GetStyleScrap

	Returns a StScrpHandle that contains all the style information for the
	current selection. This method could easily fail if the text is heavily
	styled, so the operation is pre-flighted first.
	
 ******************************************************************************/

StScrpHandle CStyleText::GetStyleScrap( void)
{
	long			styleScrapSize;
	StScrpHandle	h;

	/* pre-flight allocating memory for the style scrap */	
	
	styleScrapSize = TENumStyles( (**macTE).selStart, (**macTE).selEnd, macTE)
						* sizeof(ScrpSTElement) + 2;
						
	h = NewHandleCanFail( styleScrapSize);
	FailNIL( h);
	DisposHandle( (Handle) h);
	
	h = GetStylScrap( macTE);
	FailNIL( h);	/* shouldn't happen, but checking doesn't hurt */
	
	return h;
}

/******************************************************************************
 SetStyleScrap

	Apply a StScrpHandle to the given selection range. It is the responsibility
	of the caller to dispose of the handle.
 ******************************************************************************/

void CStyleText::SetStyleScrap( long rangeStart, long rangeEnd, StScrpHandle styleScrap,
					Boolean redraw)
{
	Prepare();
	SetStylScrap( rangeStart, rangeEnd, styleScrap, redraw, macTE);
	TECalText( macTE);
	AdjustBounds();
	
}
