/******************************************************************************
 CTextEditTask.c

		Provides undo support for typing and cut/copy/paste/clear in 
		CAbstractText subclasses.
		
		An instance of this class is automatically created by CAbstract 
		text for typing and edit operations. CEditText can use this
		class directly. Other subclasses of CAbstractText may need
		to subclass CTextEditTask. For example, CStyleText subclasses
		CTextEditTask because it must deal directly with the styled scrap 
		format used by Styled TextEdit.
		
	SUPERCLASS = CTask
	
	Copyright  1991 Symantec Corporation. All rights reserved.

	
 ******************************************************************************/

#include "CTextEditTask.h"
#include "CClipboard.h"
#include "Constants.h"
#include "Commands.h"
#include "TCLUtilities.h"
#include "CAbstractText.h"
#include "Global.h"

extern CClipboard	*gClipboard;

/******************************************************************************
 ITextEditTask

	Initialize the TextEditTask. anEditCmd should be the command that the
	task is to respond to, i.e. cmdCut, cmdCopy, cmdPaste, or cmdClear. If
	the task is for typing, then it should be cmdNull. FirstTaskIndex is 
	the index into the strUNDO 'STR#' where the text undo labels are located.
	These are typically "Typing", "Cut", "Copy", "Paste", "Clear". This method
	determines the appropriate task index to pass to CTask::ITask, determines
	whether text and or clipboard contents should be saved, and saves the
	range about to be deleted.
	
 ******************************************************************************/

void CTextEditTask::ITextEditTask( CAbstractText *aTextPane, long anEditCmd, 
						short firstTaskIndex)
{
	short	taskIndex;
	long	selStart, selEnd;
	Handle	h;
	
	itsTextPane = aTextPane;
	editCmd = anEditCmd;
	
	originalScrap = inserted.text = deleted.text = NULL;
	
	if (firstTaskIndex > 0)
	{
		if (anEditCmd == cmdNull)	/* task is Typing	*/
			taskIndex = firstTaskIndex;
		else
			taskIndex = firstTaskIndex + anEditCmd - cmdCut + 1;
	}
	else taskIndex = 0;
	
	stillTyping = anEditCmd == cmdNull;
		
	CTask::ITask( taskIndex);
	
	doClip = (editCmd == cmdCopy) || (editCmd == cmdCut);
	doText = editCmd != cmdCopy;

	itsTextPane->GetSelection( &selStart, &selEnd);
	deleted.start = deleted.selStart = selStart;
	deleted.end = deleted.selEnd = selEnd;
	inserted.start = inserted.selStart = selStart;
	inserted.end = inserted.selEnd = selEnd;
	
	SaveRange( kDeletedRange);

	if (doClip)
	{
		gClipboard->GetData( 'TEXT', &h);
		originalScrap = h;
	}
}


/******************************************************************************
 Do

	Call for the Edit menu commands. Commands the text pane to perform
	the command, and stores the resulting selection.
	
 ******************************************************************************/

void CTextEditTask::Do(void)
{
	long selStart, selEnd;
	
	itsTextPane->PerformEditCommand( editCmd);
	
	itsTextPane->GetSelection( &selStart, &selEnd);
	inserted.selStart = selStart;
	inserted.end = inserted.selEnd = selEnd;
}

/******************************************************************************
 DoTyping

	Handles keystrokes. Keystrokes are categorized into backspace, forward delete,
	and other, and a separate method is called for each. This is only called
	during the "Do" phase, never the "Undo" phase.
	
 ******************************************************************************/

void CTextEditTask::DoTyping( char theChar, short keyCode, EventRecord *macEvent)
{
	long selStart, selEnd, newLen, oldLen = itsTextPane->GetLength();

	ASSERT( undone == FALSE);
	
	typingEvent = *macEvent;
	
	if (theChar == '\b')
		DoBackspace();
	else if (keyCode == KeyFwdDelete)
		DoFwdDelete();
	else
		DoNormalChar( theChar);
		
	if (stillTyping)
	{
		newLen = itsTextPane->GetLength();
		inserted.end += newLen - oldLen;
		if (inserted.end < 0) inserted.end = 0;
	}

}

/******************************************************************************
 DoNormalChar

	Handles keystrokes that are not backspace or forward delete, typically
	characters to be inserted or cursor keys. They are passed through to the
	text pane.
 ******************************************************************************/

void CTextEditTask::DoNormalChar( char theChar)
{
	itsTextPane->TypeChar( theChar, typingEvent.modifiers);
	
}

/******************************************************************************
 DoBackspace

	Handles the backspace char. If the backspace is during the "Do" phase,
	then the text being deleted must be saved before it is gone.
	
 ******************************************************************************/

void CTextEditTask::DoBackspace( void)
{
	long 		selStart, selEnd, prevCharPos;
	tCharBuf	charBuf;
	short		i;
	long		bufSize;
	
	itsTextPane->GetSelection( &selStart, &selEnd);

	/* One of the following cases must hold:
	 *	- This is the first keystroke, and we have already
	 *	   saved the selection, or the selection is empty. In
	 *	   this case selStart == initialStart.
	 *	- We are in the middle of typing, and the selection
	 *    is guaranteed to be empty. If selStart > initialStart
	 *    then we are deleting new text and don't need to
	 *    save it. If selStart == deleted.start, we save the
	 *    character about to be deleted. selStart cannot be
	 *    less than initialStart because each backspace
	 *    decrements initialStart.
	 */

	if (selStart > inserted.start)	/* deleting new text	*/
	{
	}
	else if (selEnd > selStart)
	{
		/* must be first key and text is already saved,	*/
		/* do nothing extra								*/
	}
	else 							/* deleting original text */
	{
		prevCharPos = selStart;
		itsTextPane->GetCharBefore( &prevCharPos, charBuf);
		
		if (Length(charBuf) < 1) 	/* was there a previous char?	*/
			return;

		/* resize the saved text handle, and copy new char into place */
		
		bufSize = GetHandleSize( deleted.text);
		ResizeHandleCanFail( deleted.text, bufSize + Length(charBuf));
		FailMemError();
		
		BlockMove( *deleted.text, *deleted.text + charBuf[0], bufSize);
		for(i = 0; i < Length(charBuf); i++)
			(*deleted.text)[i] = charBuf[i+1];
						
		inserted.start -= Length(charBuf);
		deleted.start -= Length(charBuf);
	}
		
	itsTextPane->TypeChar( '\b',  typingEvent.modifiers);

}

/******************************************************************************
 DoFwdDelete

	Handles forward delete. The text that is about to be deleted must be saved
	for undo.
 ******************************************************************************/

void CTextEditTask::DoFwdDelete( void)
{
	long 		selStart, selEnd, nextCharPos;
	Handle		text;
	tCharBuf	charBuf;
	short	i;
	long	bufSize;
	
	itsTextPane->GetSelection( &selStart, &selEnd);
	
	/*
	/* One of the following cases must hold:
	 *	- This is the first keystroke, and we have already
	 *	   saved the selection, or the selection is empty. In
	 *	   this case selStart == initialStart.
	 *	- We are in the middle of typing, and the selection
	 *    is guaranteed to be empty. Forward delete
	 *	  only deletes old text.
	 */

	if (selEnd > selStart)
	{
		/* must be first key and text is already saved,	*/
		/* just clear old text							*/
		
		itsTextPane->TypeChar( '\b',  typingEvent.modifiers);
	}
	else 							/* deleting original text */
	{
		nextCharPos = selStart;
		itsTextPane->GetCharAfter( &nextCharPos, charBuf);
		
		if (Length(charBuf) < 1) 	/* was there a next char?	*/
			return;
		
		/* resize the saved text handle, and copy new char into place */
		
		bufSize = GetHandleSize( deleted.text);
		ResizeHandleCanFail( deleted.text, bufSize + Length(charBuf));
		FailMemError();
		
		for(i = 0; i < Length(charBuf); i++)
			(*deleted.text)[bufSize++] = charBuf[i+1];

		deleted.end += Length(charBuf);
		
		itsTextPane->TypeChar( kRightCursor,  typingEvent.modifiers);
		itsTextPane->TypeChar( '\b',  typingEvent.modifiers);		
	}			
}

/******************************************************************************
 Undo

	Called when the user selects Undo. Saves off any text that has been 
	inserted, deletes the inserted text, then restores any text that had
	been deleted. If needed also restores the old clipboard.
 ******************************************************************************/

void CTextEditTask::Undo(void)
{	
	inherited::Undo();
	
	if (doText)
	{
		if (stillTyping)
			CancelTyping();
			
		itsTextPane->Prepare();

		SaveRange( kInsertedRange);	
			
		/* deleted any inserted text						*/
		
		DeleteRange( kInsertedRange);
					
		/* insert any deleted text, restore selection		*/
		
		RestoreRange( kDeletedRange, FALSE);
			
	}
	
	if (doClip)		/* restore the clipboard				*/
		StoreToClip( kOldClip);
	
}

/******************************************************************************
 Redo

	Called when the user selects Redo. Re-deletes any deleted text,
	and restores any inserted text. If needed also restores the new clipboard.
 ******************************************************************************/

void CTextEditTask::Redo(void)
{
	undone = FALSE;
	if (editCmd == cmdNull) stillTyping = TRUE;
	
	if (doText)
	{		
		itsTextPane->Prepare();
		
		/* re-delete the deleted text		*/
		
		DeleteRange( kDeletedRange);
			
		/* re-insert the inserted text		*/
		
		RestoreRange( kInsertedRange, TRUE);
					
		if (editCmd == cmdNull)
			stillTyping = TRUE;	
	}
	
	if (doClip)
		StoreToClip( kNewClip);
}

/******************************************************************************
 CancelTyping

	Multiple keystrokes are are accumulated into a single undoable action.
	CancelTyping is called when editing must be stopped, such as when the
	user changes the selection area or selects Undo.
	
 ******************************************************************************/

void CTextEditTask::CancelTyping(void)
{
	long selStart, selEnd;
	
	stillTyping = FALSE;	
	itsTextPane->GetSelection( &selStart, &selEnd);
	inserted.selStart = selStart;
	inserted.selEnd = selEnd;
}

/******************************************************************************
 CanStillType

	Returns TRUE if the task can still accumulate more keystrokes for undo. The
	task must be in the "Do" and not "Undo" phase, and stillTyping must be TRUE.
 ******************************************************************************/

Boolean CTextEditTask::CanStillType(void)
{
	return stillTyping && !undone;
}

/******************************************************************************
 SelectionChanged

	Called by the text pane when the selection changes.
 ******************************************************************************/

void CTextEditTask::SelectionChanged(void)
{
	if (!undone)
		CancelTyping();
}

/******************************************************************************
 Dispose

	Dispose of memory for this object. This method also clears the references
	that the text pane keeps to this task.
 ******************************************************************************/

void CTextEditTask::Dispose(void)
{
	ForgetHandle( originalScrap);
	ForgetHandle( inserted.text);
	ForgetHandle( deleted.text);
	if (itsTextPane)
	{
		if (itsTextPane->itsTypingTask == this)
			itsTextPane->itsTypingTask = NULL;
			
		if (itsTextPane->itsLastTask == this)
			itsTextPane->itsLastTask = NULL;
	}
	inherited::Dispose();
}

/******************************************************************************
 SaveRange

	Saves either the text of the inserted or deleted range.
 ******************************************************************************/

void CTextEditTask::SaveRange( tRangeSelector whichRange)
{
	Handle	h = NULL;
	
	if (whichRange == kInsertedRange)
	{
		h = itsTextPane->CopyTextRange( inserted.start, inserted.end);
		inserted.text = h;
	}
	else
	{
		h = itsTextPane->CopyTextRange( deleted.start, deleted.end);
		deleted.text = h;
	}
}

/******************************************************************************
 DeleteRange

	Deletes the text of the inserted or deleted range.
 ******************************************************************************/

void CTextEditTask::DeleteRange( tRangeSelector whichRange)
{
	if (whichRange == kInsertedRange)
		itsTextPane->SetSelection( inserted.start, inserted.end, TRUE);
	else
		itsTextPane->SetSelection( deleted.start, deleted.end, TRUE);
		
	itsTextPane->PerformEditCommand( cmdClear);
}

/******************************************************************************
 RestoreRange

	Restores the text of either the inserted or deleted range. If killData
	is TRUE, the saved text is disposed.
 ******************************************************************************/

void CTextEditTask::RestoreRange( tRangeSelector whichRange, Boolean killData)
{
	Handle h = NULL;
	
	if (whichRange == kInsertedRange)
	{
		h = inserted.text;
		itsTextPane->SetSelection( inserted.start, inserted.start, TRUE);
		itsTextPane->InsertTextHandle( h, FALSE);	
		itsTextPane->SetSelection( inserted.selStart, inserted.selEnd, TRUE);
		if (killData) inserted.text = NULL;	
	}
	else
	{
		h = deleted.text;
		itsTextPane->SetSelection( deleted.start, deleted.start, TRUE);
		itsTextPane->InsertTextHandle( h, FALSE);		
		itsTextPane->SetSelection( deleted.selStart, deleted.selEnd, TRUE);		
		if (killData) deleted.text = NULL;
	}
	if (killData)
		ForgetHandle(h);
}

/******************************************************************************
 StoreToClip

	Stores either the original scrap text or the deleted text to
	the clipboard.
 ******************************************************************************/

void CTextEditTask::StoreToClip( tClipSelector whichClip)
{
	Handle h;
	
	if (whichClip == kOldClip)
		h = originalScrap;
	else
		h = deleted.text;

	gClipboard->EmptyScrap();
	gClipboard->PutData( 'TEXT', h);
}
